import os
from dataclasses import dataclass, replace, field, fields
import dis
import operator
from functools import reduce
from typing import (
    Optional,
    Protocol,
    runtime_checkable,
    Union,
    NamedTuple,
    Mapping,
    TypeVar,
    TYPE_CHECKING,
)

from collections import ChainMap
from numba_rvsdg.core.datastructures.byte_flow import ByteFlow
from numba_rvsdg.core.datastructures.scfg import SCFG
from numba_rvsdg.core.datastructures.basic_block import (
    BasicBlock,
    PythonBytecodeBlock,
    RegionBlock,
    SyntheticBranch,
    SyntheticAssignment,
)
from numba_rvsdg.rendering.rendering import ByteFlowRenderer
from numba_rvsdg.core.datastructures import block_names


from numba.core.utils import MutableSortedSet, MutableSortedMap

from .regionpasses import (
    RegionVisitor,
    RegionTransformer,
)

if TYPE_CHECKING:
    from .regionrenderer import GraphBuilder


DEBUG_GRAPH = int(os.environ.get("DEBUG_GRAPH", "0"))


T = TypeVar("T")


def _just(v: Optional[T]) -> T:
    """Unpack optional type."""
    assert v is not None
    return v


@dataclass(frozen=True)
class ValueState:
    """For representing a RVSDG Value (State).

    For most compiler passes, Value and State can be treated as the same.
    """
    parent: Optional["Op"]
    """Optional. The parent Op that output this ValueState.
    """
    name: str
    """Name of the Value(State).
    """
    out_index: int
    """The output port index in the parent Op.
    """
    is_effect: bool = False
    """True if-and-only-if this is a state.
    """

    def short_identity(self) -> str:
        args = f"{id(self.parent):x}, {self.name}, {self.out_index}"
        return f"ValueState({args})"

    def __hash__(self):
        return id(self)


@dataclass(frozen=True)
class Op:
    """For representing a RVSDG operation.

    An Op has inputs and outputs ports that are ValueStates.
    An Op can optionally have a bytecode instruction associated with it.
    """
    opname: str
    """Operation name.
    """
    bc_inst: Optional[dis.Instruction]
    """Optional. The bytecode instruction.
    """
    _inputs: dict[str, ValueState] = field(default_factory=dict)
    """The input ports.
    """
    _outputs: dict[str, ValueState] = field(default_factory=dict)
    """The output ports.
    """

    def add_input(self, name, vs: ValueState):
        self._inputs[name] = vs

    def add_output(self, name: str, is_effect=False) -> ValueState:
        vs = ValueState(
            parent=self,
            name=name,
            out_index=len(self._outputs),
            is_effect=is_effect,
        )
        self._outputs[name] = vs
        return vs

    def short_identity(self) -> str:
        return f"Op({self.opname}, {id(self):x})"

    def summary(self) -> str:
        ins = ", ".join([k for k in self._inputs])
        outs = ", ".join([k for k in self._outputs])
        bc = "---"
        if self.bc_inst is not None:
            bc = f"{self.bc_inst.opname}({self.bc_inst.argrepr})"
        return f"Op\n{self.opname}\n{bc}\n({ins}) -> ({outs}) "

    @property
    def outputs(self) -> list[ValueState]:
        return list(self._outputs.values())

    @property
    def inputs(self) -> list[ValueState]:
        return list(self._inputs.values())

    @property
    def input_ports(self) -> Mapping[str, ValueState]:
        return self._inputs

    @property
    def output_ports(self) -> Mapping[str, ValueState]:
        return self._outputs

    def __hash__(self):
        return id(self)


@runtime_checkable
class DDGProtocol(Protocol):
    incoming_states: MutableSortedSet[str]
    outgoing_states: MutableSortedSet[str]


@dataclass(frozen=True)
class DDGRegion(RegionBlock):
    incoming_states: MutableSortedSet[str] = field(
        default_factory=MutableSortedSet
    )
    outgoing_states: MutableSortedSet[str] = field(
        default_factory=MutableSortedSet
    )


@dataclass(frozen=True)
class DDGBranch(SyntheticBranch):
    incoming_states: MutableSortedSet[str] = field(
        default_factory=MutableSortedSet
    )
    outgoing_states: MutableSortedSet[str] = field(
        default_factory=MutableSortedSet
    )


@dataclass(frozen=True)
class DDGControlVariable(SyntheticAssignment):
    incoming_states: MutableSortedSet[str] = field(
        default_factory=MutableSortedSet
    )
    outgoing_states: MutableSortedSet[str] = field(
        default_factory=MutableSortedSet
    )


@dataclass(frozen=True)
class DDGBlock(BasicBlock):
    in_effect: ValueState | None = None
    out_effect: ValueState | None = None
    in_stackvars: list[ValueState] = field(default_factory=list)
    out_stackvars: list[ValueState] = field(default_factory=list)
    in_vars: MutableSortedMap[str, ValueState] = field(
        default_factory=MutableSortedMap
    )
    out_vars: MutableSortedMap[str, ValueState] = field(
        default_factory=MutableSortedMap
    )

    exported_stackvars: MutableSortedMap[str, ValueState] = field(
        default_factory=MutableSortedMap
    )

    def __post_init__(self):
        assert isinstance(self.in_vars, MutableSortedMap)
        assert isinstance(self.out_vars, MutableSortedMap)

    def _gather_reachable(
        self, vs: ValueState, reached: set[ValueState]
    ) -> set[ValueState]:
        reached.add(vs)
        if vs.parent is not None:
            for ivs in vs.parent.inputs:
                if ivs not in reached:
                    self._gather_reachable(ivs, reached)
        return reached

    def render_graph(self, builder: "GraphBuilder"):
        reached_vs: set[ValueState] = set()
        for vs in [*self.out_vars.values(), _just(self.out_effect)]:
            self._gather_reachable(vs, reached_vs)
        reached_vs.add(_just(self.in_effect))
        reached_vs.update(self.in_vars.values())

        reached_op = {vs.parent for vs in reached_vs if vs.parent is not None}

        for vs in reached_vs:
            self._render_vs(builder, vs)

        for op in reached_op:
            self._render_op(builder, op)

        # Make outgoing node
        ports = []
        outgoing_nodename = f"outgoing_{self.name}"
        for k, vs in self.out_vars.items():
            ports.append(k)

            builder.graph.add_edge(
                vs.short_identity(), outgoing_nodename, dst_port=k
            )

        outgoing_node = builder.node_maker.make_node(
            kind="ports",
            ports=ports,
            data=dict(body="outgoing"),
        )
        builder.graph.add_node(outgoing_nodename, outgoing_node)

        # Make incoming node
        ports = []
        incoming_nodename = f"incoming_{self.name}"
        for k, vs in self.in_vars.items():
            ports.append(k)

            builder.graph.add_edge(
                incoming_nodename,
                _just(vs.parent).short_identity(),
                src_port=k,
            )

        incoming_node = builder.node_maker.make_node(
            kind="ports",
            ports=ports,
            data=dict(body="incoming"),
        )
        builder.graph.add_node(incoming_nodename, incoming_node)

        # Make jump target node for debugging
        jt_node = builder.node_maker.make_node(
            kind="meta",
            data=dict(body=f"jump-targets: {self._jump_targets}"),
        )
        builder.graph.add_node(f"jt_{self.name}", jt_node)

    def _render_vs(self, builder: "GraphBuilder", vs: ValueState):
        if vs.is_effect:
            node = builder.node_maker.make_node(
                kind="effect",
                data=dict(body=str(vs.name)),
            )
            builder.graph.add_node(vs.short_identity(), node)
        else:
            node = builder.node_maker.make_node(
                kind="valuestate",
                data=dict(body=str(vs.name)),
            )
            builder.graph.add_node(vs.short_identity(), node)

    def _render_op(self, builder, op: Op):
        op_anchor = op.short_identity()

        node = builder.node_maker.make_node(
            kind="op",
            data=dict(body=str(op.summary())),
        )
        builder.graph.add_node(op_anchor, node)

        # draw edge
        for edgename, vs in op._outputs.items():
            self._add_vs_edge(builder, op_anchor, vs, taillabel=f"{edgename}")
        for edgename, vs in op._inputs.items():
            self._add_vs_edge(builder, vs, op_anchor, headlabel=f"{edgename}")

    def _add_vs_edge(self, builder, src, dst, **attrs):
        is_effect = (isinstance(src, ValueState) and src.is_effect) or (
            isinstance(dst, ValueState) and dst.is_effect
        )
        if isinstance(src, ValueState):
            src = src.short_identity()
        if isinstance(dst, ValueState):
            dst = dst.short_identity()

        kwargs = attrs
        if is_effect:
            kwargs["kind"] = "effect"

        builder.graph.add_edge(src, dst, **kwargs)

    @property
    def incoming_states(self) -> MutableSortedSet:
        return MutableSortedSet(self.in_vars)

    @property
    def outgoing_states(self) -> MutableSortedSet:
        return MutableSortedSet(self.out_vars)

    def get_toposorted_ops(self) -> list[Op]:
        """Get a topologically sorted list of ``Op`` according
        to the data-dependence.

        Operations stored later in the list may depend on earlier operations,
        but the reverse can never be true.
        """
        res: list[Op] = []

        # Initially, the available states are the inputs
        avail: set[ValueState] = {*self.in_vars.values(), _just(self.in_effect)}
        # Seed the pending Operations using the outputs.
        pending: list[Op] = [
            vs.parent for vs in self.out_vars.values() if vs.parent is not None
        ]
        assert self.out_effect is not None  # for typing
        pending.append(_just(self.out_effect.parent))
        seen: set[Op] = set()

        while pending:
            op = pending[-1]
            if op in seen:
                pending.pop()
                continue
            # Get the set of incoming states that are not yet available
            # to the current operation.
            # NOTE: for stable ordering, change the following to a ordered-set.
            incomings = set()
            for vs in op._inputs.values():
                if vs not in avail and vs.parent is not None:
                    incomings.add(vs.parent)

            if not incomings:
                # All incoming states are already available
                avail |= set(op._outputs.values())
                pending.pop()
                res.append(op)
                seen.add(op)
            else:
                pending.extend(incomings)
        return res


def render_scfg(byteflow):
    bfr = ByteFlowRenderer()
    bfr.bcmap_from_bytecode(byteflow.bc)
    byteflow.scfg.view("scfg")


def _repl_jump_targets(block: BasicBlock, repl: dict[str, str]):
    if set(repl).intersection(set(block.jump_targets)):
        targets = [repl.get(k, k) for k in block.jump_targets]
        block = block.replace_jump_targets(tuple(targets))
    return block


def _canonicalize_scfg_switch(scfg: SCFG):
    """Introduce "switch" region to enclose "head", "branch", "tail" regions."""
    todos = set(scfg.graph)
    while todos:
        label = todos.pop()
        todos.discard(label)
        block = scfg[label]
        if isinstance(block, RegionBlock):
            if block.kind == "head":
                brlabels = block.jump_targets
                branches = [scfg[brlabel] for brlabel in brlabels]
                tail_label_candidates = set()
                for br in branches:
                    tail_label_candidates.update(br.jump_targets)
                [taillabel] = tail_label_candidates
                tail = scfg[taillabel]

                # Make new SCFG
                switch_labels = {label, taillabel, *brlabels}
                subregion_graph = {k: scfg[k] for k in switch_labels}
                scfg.remove_blocks(switch_labels)
                subregion_scfg = SCFG(
                    graph=subregion_graph, name_gen=scfg.name_gen
                )

                todos -= switch_labels

                # Make new region
                new_label = scfg.name_gen.new_region_name("switch")
                new_region = RegionBlock(
                    name=new_label,
                    _jump_targets=tail._jump_targets,
                    kind="switch",
                    parent_region=block.parent_region,
                    header=block.name,
                    exiting=taillabel,
                    subregion=subregion_scfg,
                )
                scfg.graph[new_label] = new_region
                # fixup branch to label
                for incoming_label, incoming_blk in scfg.graph.items():
                    if (
                        incoming_label != new_label
                        and label in incoming_blk.jump_targets
                    ):
                        repl = {label: new_label}
                        replblk = _repl_jump_targets(incoming_blk, repl)
                        scfg.graph[incoming_label] = replblk

                # fixup header
                if block.parent_region.header not in scfg.graph:
                    block.parent_region.replace_header(new_label)
                # fixup exiting
                if block.parent_region.exiting not in scfg.graph:
                    block.parent_region.replace_exiting(new_label)

                # recursively walk into the subregions
                _canonicalize_scfg_switch(subregion_graph[label].subregion)
                for br in brlabels:
                    _canonicalize_scfg_switch(subregion_graph[br].subregion)
                _canonicalize_scfg_switch(subregion_graph[taillabel].subregion)
            elif block.kind == "loop":
                _canonicalize_scfg_switch(block.subregion)


class CanonicalizeLoop(RegionTransformer[None]):
    """
    Make sure loops has non-region header.

    Preferably, the exiting block should be non-region as well but
    it's hard to do with the current numba_rvsdg API.

    Doing this so we don't have to fixup backedges as backedges will always
    point to a non-region node in ``_canonicalize_scfg_switch``.
    """

    def visit_loop(self, parent: SCFG, region: RegionBlock, data: None):
        # Fix header
        # Introduce a SyntheticFill block that just jump to the original header
        new_label = parent.name_gen.new_block_name(block_names.SYNTH_FILL)
        region.subregion.insert_SyntheticFill(new_label, {}, {region.header})
        region.replace_header(new_label)

        # Fix exiting block's backedges to point to the new header
        def get_inner_most_exiting(blk):
            while isinstance(blk, RegionBlock):
                parent, blk = blk, blk.subregion.graph[blk.exiting]
            return parent, blk

        tail_parent, tail_bb = get_inner_most_exiting(region)
        [backedge] = tail_bb.backedges
        repl = {backedge: new_label}
        new_tail_bb = replace(
            tail_bb,
            backedges=(new_label,),
            _jump_targets=tuple(
                [repl.get(x, x) for x in tail_bb._jump_targets]
            ),
        )
        tail_parent.subregion.graph[tail_bb.name] = new_tail_bb

        self.visit_linear(parent, region, data)

    def visit_block(self, parent: SCFG, block: BasicBlock, data: None):
        pass  # no-op

    def visit_switch(self, parent: SCFG, block: BasicBlock, data: None):
        pass  # no-op


def canonicalize_scfg(scfg: SCFG):
    CanonicalizeLoop().visit_graph(scfg, None)
    _canonicalize_scfg_switch(scfg)


class _ExtraBranch(NamedTuple):
    branch_instlists: tuple[tuple[str, ...], ...] = ()


@dataclass(frozen=True)
class ExtraBasicBlock(BasicBlock):
    inst_list: tuple[str, ...] = ()

    @classmethod
    def make(cls, label, jump_target, instlist):
        return ExtraBasicBlock(label, (jump_target,), inst_list=instlist)

    def __str__(self):
        args = "\n".join(f"{inst})" for inst in self.inst_list)
        return f"ExtraBasicBlock({args})"


class HandleConditionalPop:
    """Introduce pop-stack operations to the bytecode to correctly model
    operations that conditionally pop elements from the stack. Numba-rvsdg does
    not handle this. For example, FOR_ITER pop the stack when the iterator is
    exhausted.
    """

    def handle(self, inst: dis.Instruction) -> _ExtraBranch | None:
        fn = getattr(self, f"op_{inst.opname}", self._op_default)
        return fn(inst)

    def _op_default(self, inst: dis.Instruction) -> None:
        # Being defensive
        assert not inst.opname.endswith("OR_POP")
        return

    def op_FOR_ITER(self, inst: dis.Instruction) -> _ExtraBranch:
        # Bytecode semantic pop 1 for the iterator
        # but we need an extra pop because the indvar is pushed
        br0 = ("FOR_ITER_STORE_INDVAR",)
        br1 = ("POP",)
        return _ExtraBranch((br0, br1))

    def op_JUMP_IF_TRUE_OR_POP(self, inst: dis.Instruction) -> _ExtraBranch:
        br0 = ("POP",)
        br1 = ()
        return _ExtraBranch((br0, br1))

    def op_JUMP_IF_FALSE_OR_POP(self, inst: dis.Instruction) -> _ExtraBranch:
        br0 = ("POP",)
        br1 = ()
        return _ExtraBranch((br0, br1))


def _scfg_add_conditional_pop_stack(bcmap, scfg: SCFG):
    extra_records = {}
    for blk in scfg.graph.values():
        if isinstance(blk, PythonBytecodeBlock):
            # Check if last instruction has conditional pop
            last_inst = blk.get_instructions(bcmap)[-1]
            handler = HandleConditionalPop()
            res = handler.handle(last_inst)
            if res is not None:
                for br_index, instlist in enumerate(res.branch_instlists):
                    k = blk._jump_targets[br_index]
                    extra_records[k] = blk.name, (br_index, instlist)

    def _replace_jump_targets(blk, idx, repl):
        return replace(
            blk,
            _jump_targets=tuple(
                [
                    repl if i == idx else jt
                    for i, jt in enumerate(blk._jump_targets)
                ]
            ),
        )

    for label, (parent_label, (br_index, instlist)) in extra_records.items():
        newlabel = scfg.name_gen.new_block_name("python.extrabasicblock")
        scfg.graph[parent_label] = _replace_jump_targets(
            scfg.graph[parent_label], br_index, newlabel
        )
        ebb = ExtraBasicBlock.make(newlabel, label, instlist)
        scfg.graph[newlabel] = ebb


def build_rvsdg(code, argnames: tuple[str, ...]) -> SCFG:
    byteflow = ByteFlow.from_bytecode(code)
    bcmap = byteflow.scfg.bcmap_from_bytecode(byteflow.bc)
    _scfg_add_conditional_pop_stack(bcmap, byteflow.scfg)
    byteflow = byteflow.restructure()
    # if DEBUG_GRAPH:
    #     render_scfg(byteflow)
    canonicalize_scfg(byteflow.scfg)
    if DEBUG_GRAPH:
        render_scfg(byteflow)
    rvsdg = convert_to_dataflow(byteflow, argnames)
    rvsdg = propagate_states(rvsdg)
    if DEBUG_GRAPH:
        from .regionrenderer import RVSDGRenderer, to_graphviz

        to_graphviz(RVSDGRenderer().render(rvsdg)).view("rvsdg")

    return rvsdg


def _flatten_full_graph(scfg: SCFG):
    regions = [
        _flatten_full_graph(elem.subregion)
        for elem in scfg.graph.values()
        if isinstance(elem, RegionBlock)
    ]
    out = ChainMap(*regions, scfg.graph)
    for blk in out.values():
        assert not isinstance(blk, RegionBlock), type(blk)
    return out


DDGTypes = (DDGBlock, DDGControlVariable, DDGBranch)
_DDGTypeAnn = Union[DDGBlock, DDGControlVariable, DDGBranch]


def convert_to_dataflow(byteflow: ByteFlow, argnames: tuple[str, ...]) -> SCFG:
    bcmap = {inst.offset: inst for inst in byteflow.bc}
    rvsdg = convert_scfg_to_dataflow(byteflow.scfg, bcmap, argnames)
    return rvsdg


def propagate_states(rvsdg: SCFG) -> SCFG:
    propagate_stack(rvsdg)
    propagate_vars(rvsdg)
    connect_incoming_stack_vars(rvsdg)
    return rvsdg


def propagate_vars(rvsdg: SCFG):
    # Propagate variables
    visitor = PropagateVars()
    visitor.visit_graph(rvsdg, visitor.make_data())


_pvData = set[str]


class PropagateVars(RegionVisitor[_pvData]):
    """
    Depends on PropagateStack
    """

    def __init__(self, _debug: bool = False):
        super().__init__()
        self._debug = _debug

    def debug_print(self, *args, **kwargs):
        if self._debug:
            print(*args, **kwargs)

    def _apply(self, block: BasicBlock, data: _pvData) -> _pvData:
        assert isinstance(block, BasicBlock)
        if isinstance(block, DDGProtocol):
            if isinstance(block, DDGBlock):
                for k in data:
                    if k not in block.in_vars:
                        op = Op(opname="var.incoming", bc_inst=None)
                        vs = op.add_output(k)
                        block.in_vars[k] = vs
                        if k.startswith("tos."):
                            if k in block.exported_stackvars:
                                block.out_vars[k] = block.exported_stackvars[k]
                        elif k not in block.out_vars:
                            block.out_vars[k] = vs
            else:
                block.incoming_states.update(data)
                block.outgoing_states.update(data)

            data = set(block.outgoing_states)
            return data
        else:
            return data

    def visit_linear(self, region: RegionBlock, data: _pvData) -> _pvData:
        region.incoming_states.update(data)
        data = self.visit_graph(region.subregion, data)
        region.outgoing_states.update(data)
        return set(region.outgoing_states)

    def visit_block(self, block: BasicBlock, data: _pvData) -> _pvData:
        return self._apply(block, data)

    def visit_loop(self, region: RegionBlock, data: _pvData) -> _pvData:
        self.debug_print("---LOOP_ENTER", region.name, data)
        data = self.visit_linear(region, data)
        self.debug_print("---LOOP_END=", region.name, "vars", data)
        return data

    def visit_switch(self, region: RegionBlock, data: _pvData) -> _pvData:
        self.debug_print("---SWITCH_ENTER", region.name)
        region.incoming_states.update(data)
        header = region.header
        data_at_head = self.visit_linear(region.subregion[header], data)
        data_for_branches = []
        for blk in region.subregion.graph.values():
            if blk.kind == "branch":
                data_for_branches.append(self.visit_linear(blk, data_at_head))
        data_after_branches = reduce(operator.or_, data_for_branches)

        exiting = region.exiting

        data_at_tail = self.visit_linear(
            region.subregion[exiting], data_after_branches
        )

        self.debug_print("data_at_head", data_at_head)
        self.debug_print("data_for_branches", data_for_branches)
        self.debug_print("data_after_branches", data_after_branches)
        self.debug_print("data_at_tail", data_at_tail)
        self.debug_print("---SWITCH_END=", region.name, "vars", data_at_tail)
        region.outgoing_states.update(data_at_tail)
        return set(region.outgoing_states)

    def make_data(self) -> _pvData:
        return set()


_psData = tuple[str, ...]


class PropagateStack(RegionVisitor[_psData]):
    def __init__(self, _debug: bool = False):
        super(PropagateStack, self).__init__()
        self._debug = _debug

    def debug_print(self, *args, **kwargs):
        if self._debug:
            print(*args, **kwargs)

    def visit_block(self, block: BasicBlock, data: _psData) -> _psData:
        if isinstance(block, DDGBlock):
            nin = len(block.in_stackvars)
            inherited = data[: len(data) - nin]
            self.debug_print("--- stack", data)
            self.debug_print("--- inherited stack", inherited)
            out_stackvars = block.out_stackvars.copy()

            counts = reversed(list(range(len(inherited) + len(out_stackvars))))
            out_stack = block.out_stackvars[::-1]
            out_data = tuple([f"tos.{i}" for i in counts])

            unused_names = list(out_data)

            for vs in reversed(out_stack):
                k = unused_names.pop()
                op = Op("stack.export", bc_inst=None)
                op.add_input("0", vs)
                block.exported_stackvars[k] = vs = op.add_output(k)
                block.out_vars[k] = vs

            for orig in reversed(inherited):
                k = unused_names.pop()
                import_op = Op("var.incoming", bc_inst=None)
                block.in_vars[orig] = imported_vs = import_op.add_output(orig)

                op = Op("stack.export", bc_inst=None)
                op.add_input("0", imported_vs)
                vs = op.add_output(k)
                block.exported_stackvars[k] = vs
                block.out_vars[k] = vs

            self.debug_print("---=", block.name, "out stack", out_data)
            return out_data
        else:
            return data

    def visit_loop(self, region: RegionBlock, data: _psData) -> _psData:
        self.debug_print("---LOOP_ENTER", region.name)
        data = self.visit_linear(region, data)
        self.debug_print("---LOOP_END=", region.name, "stack", data)
        return data

    def visit_switch(self, region: RegionBlock, data: _psData) -> _psData:
        self.debug_print("---SWITCH_ENTER", region.name)
        header = region.header
        data_at_head = self.visit_linear(region.subregion[header], data)
        data_for_branches = []
        for blk in region.subregion.graph.values():
            if blk.kind == "branch":
                data_for_branches.append(self.visit_linear(blk, data_at_head))
        data_after_branches = max(data_for_branches, key=len)

        exiting = region.exiting

        data_at_tail = self.visit_linear(
            region.subregion[exiting], data_after_branches
        )

        self.debug_print("data_at_head", data_at_head)
        self.debug_print("data_for_branches", data_for_branches)
        self.debug_print("data_after_branches", data_after_branches)
        self.debug_print("data_at_tail", data_at_tail)
        self.debug_print("---SWITCH_END=", region.name, "stack", data_at_tail)
        return data_at_tail

    def make_data(self) -> _psData:
        # Stack data stored in a tuple
        return ()


class ConnectImportedStackVars(RegionVisitor[None]):
    def visit_block(self, block: BasicBlock, data: None):
        if isinstance(block, DDGBlock):
            # Connect stack.incoming node to the import stack variable.
            imported_stackvars = [
                var for k, var in block.in_vars.items() if k.startswith("tos.")
            ][::-1]
            n = len(block.in_stackvars)
            for vs, inc in zip(block.in_stackvars, imported_stackvars[-n:]):
                assert vs.parent is not None
                vs.parent.add_input("0", inc)

    def visit_loop(self, region: RegionBlock, data: None):
        self.visit_linear(region, data)

    def visit_switch(self, region: RegionBlock, data: None):
        header = region.header
        self.visit_linear(region.subregion[header], data)
        for blk in region.subregion.graph.values():
            if blk.kind == "branch":
                self.visit_linear(blk, None)
        exiting = region.exiting
        self.visit_linear(region.subregion[exiting], None)


def propagate_stack(rvsdg: SCFG):
    visitor = PropagateStack(_debug=False)
    visitor.visit_graph(rvsdg, visitor.make_data())


def connect_incoming_stack_vars(rvsdg: SCFG):
    ConnectImportedStackVars().visit_graph(rvsdg, None)


def _upgrade_dataclass(old, newcls, replacements=None):
    if replacements is None:
        replacements = {}
    fieldnames = [fd.name for fd in fields(old)]
    oldattrs = {k: getattr(old, k) for k in fieldnames if k not in replacements}
    return newcls(**oldattrs, **replacements)


def convert_scfg_to_dataflow(scfg, bcmap, argnames: tuple[str, ...]) -> SCFG:
    rvsdg = SCFG()
    for block in scfg.graph.values():
        # convert block
        if isinstance(block, PythonBytecodeBlock):
            ddg = convert_bc_to_ddg(block, bcmap, argnames)
            rvsdg.add_block(ddg)
        elif isinstance(block, RegionBlock):
            # Inside-out
            subregion = convert_scfg_to_dataflow(
                block.subregion, bcmap, argnames
            )
            rvsdg.add_block(
                _upgrade_dataclass(block, DDGRegion, dict(subregion=subregion))
            )
        elif isinstance(block, SyntheticBranch):
            rvsdg.add_block(_upgrade_dataclass(block, DDGBranch))
        elif isinstance(block, SyntheticAssignment):
            rvsdg.add_block(_upgrade_dataclass(block, DDGControlVariable))
        elif isinstance(block, ExtraBasicBlock):
            ddg = convert_extra_bb(block)
            rvsdg.add_block(ddg)
        elif isinstance(block, BasicBlock):
            start_env = Op("start", bc_inst=None)
            effect = start_env.add_output("env", is_effect=True)
            newblk = _upgrade_dataclass(
                block, DDGBlock, dict(in_effect=effect, out_effect=effect)
            )
            rvsdg.add_block(newblk)
        else:
            raise Exception("unreachable", type(block))

    return rvsdg


def _convert_bytecode(block, instlist, argnames: tuple[str, ...]) -> DDGBlock:
    converter = BC2DDG()
    if instlist[0].offset == 0:
        for arg in argnames:
            converter.load(f"var.{arg}")
    for inst in instlist:
        converter.convert(inst)
    return _converter_to_ddgblock(block, converter)


def _converter_to_ddgblock(block, converter) -> DDGBlock:
    blk = DDGBlock(
        name=block.name,
        _jump_targets=block._jump_targets,
        backedges=block.backedges,
        in_effect=converter.in_effect,
        out_effect=converter.effect,
        in_stackvars=list(converter.incoming_stackvars),
        out_stackvars=list(converter.stack),
        in_vars=MutableSortedMap(converter.incoming_vars),
        out_vars=MutableSortedMap(converter.varmap),
    )
    return blk


def convert_extra_bb(block: ExtraBasicBlock) -> DDGBlock:
    converter = BC2DDG()
    for opname in block.inst_list:
        if opname == "FOR_ITER_STORE_INDVAR":
            converter.push(converter.load("indvar"))
        elif opname == "POP":
            converter.pop()
        else:
            assert False, opname
    return _converter_to_ddgblock(block, converter)


def convert_bc_to_ddg(
    block: PythonBytecodeBlock,
    bcmap: dict[int, dis.Bytecode],
    argnames: tuple[str, ...],
) -> DDGBlock:
    instlist = block.get_instructions(bcmap)
    return _convert_bytecode(block, instlist, argnames)


class BC2DDG:
    stack: list[ValueState]
    effect: ValueState
    in_effect: ValueState
    varmap: dict[str, ValueState]
    incoming_vars: dict[str, ValueState]
    incoming_stackvars: list[ValueState]
    _kw_names: ValueState | None

    def __init__(self):
        self.stack = []
        start_env = Op("start", bc_inst=None)
        self.effect = start_env.add_output("env", is_effect=True)
        self.in_effect = self.effect
        self.varmap = {}
        self.incoming_vars = {}
        self.incoming_stackvars = []
        self._kw_names = None

    def push(self, val: ValueState):
        self.stack.append(val)

    def pop(self) -> ValueState:
        if not self.stack:
            op = Op(opname="stack.incoming", bc_inst=None)
            vs = op.add_output(f"stack.{len(self.incoming_stackvars)}")
            self.stack.append(vs)
            self.incoming_stackvars.append(vs)
        return self.stack.pop()

    def top(self) -> ValueState:
        tos = self.pop()
        self.push(tos)
        return tos

    def _decorate_varname(self, varname: str) -> str:
        return f"var.{varname}"

    def store(self, varname: str, value: ValueState):
        self.varmap[varname] = value

    def load(self, varname: str) -> ValueState:
        if varname not in self.varmap:
            op = Op(opname="var.incoming", bc_inst=None)
            vs = op.add_output(varname)
            self.incoming_vars[varname] = vs
            self.varmap[varname] = vs

        return self.varmap[varname]

    def replace_effect(self, env: ValueState):
        assert env.is_effect
        self.effect = env

    def convert(self, inst: dis.Instruction):
        fn = getattr(self, f"op_{inst.opname}")
        fn(inst)

    def set_kw_names(self, kw_vs: ValueState):
        assert self._kw_names is None
        self._kw_names = kw_vs

    def pop_kw_names(self):
        res = self._kw_names
        self._kw_names = None
        return res

    def op_POP_TOP(self, inst: dis.Instruction):
        self.pop()

    def op_RESUME(self, inst: dis.Instruction):
        pass  # no-op

    def op_COPY_FREE_VARS(self, inst: dis.Instruction):
        pass  # no-op

    def op_PUSH_NULL(self, inst: dis.Instruction):
        op = Op(opname="push_null", bc_inst=inst)
        null = op.add_output("null")
        self.push(null)

    def op_LOAD_GLOBAL(self, inst: dis.Instruction):
        assert isinstance(inst.arg, int)  # for typing
        load_null = inst.arg & 1
        op = Op(opname="global", bc_inst=inst)
        op.add_input("env", self.effect)
        null = op.add_output("null")
        if load_null:
            self.push(null)
        self.push(op.add_output(f"{inst.argval}"))

    def op_LOAD_CONST(self, inst: dis.Instruction):
        op = Op(opname="const", bc_inst=inst)
        self.push(op.add_output("out"))

    def op_STORE_FAST(self, inst: dis.Instruction):
        tos = self.pop()
        op = Op(opname="store", bc_inst=inst)
        op.add_input("value", tos)
        varname = self._decorate_varname(inst.argval)
        self.store(varname, op.add_output(varname))

    def op_LOAD_FAST(self, inst: dis.Instruction):
        varname = self._decorate_varname(inst.argval)
        self.push(self.load(varname))

    def op_LOAD_ATTR(self, inst: dis.Instruction):
        obj = self.pop()
        attr = inst.argval
        op = Op(opname=f"load_attr.{attr}", bc_inst=inst)
        op.add_input("obj", obj)
        self.push(op.add_output("out"))

    def op_LOAD_METHOD(self, inst: dis.Instruction):
        obj = self.pop()
        attr = inst.argval
        op = Op(opname=f"load_method.{attr}", bc_inst=inst)
        op.add_input("obj", obj)
        self.push(op.add_output("null"))
        self.push(op.add_output("out"))

    def op_LOAD_DEREF(self, inst: dis.Instruction):
        op = Op(opname="load_deref", bc_inst=inst)
        self.push(op.add_output("out"))

    def op_PRECALL(self, inst: dis.Instruction):
        pass  # no-op

    def op_KW_NAMES(self, inst: dis.Instruction):
        op = Op(opname="kw_names", bc_inst=inst)
        self.set_kw_names(op.add_output("out"))

    def op_CALL(self, inst: dis.Instruction):
        argc: int = inst.argval
        arg1plus = reversed([self.pop() for _ in range(argc)])
        arg0 = self.pop()  # TODO
        kw_names = self.pop_kw_names()

        args: list[ValueState] = [arg0, *arg1plus]
        callable = self.pop()  # TODO
        opname = "call" if kw_names is None else "call.kw"
        op = Op(opname=opname, bc_inst=inst)
        op.add_input("env", self.effect)
        op.add_input("callee", callable)
        for i, arg in enumerate(args):
            op.add_input(f"arg.{i}", arg)

        if kw_names is not None:
            op.add_input("kw_names", kw_names)

        self.replace_effect(op.add_output("env", is_effect=True))
        self.push(op.add_output("ret"))

    def op_GET_ITER(self, inst: dis.Instruction):
        tos = self.pop()
        op = Op(opname="getiter", bc_inst=inst)
        op.add_input("obj", tos)
        self.push(op.add_output("iter"))

    def op_FOR_ITER(self, inst: dis.Instruction):
        tos = self.top()
        op = Op(opname="foriter", bc_inst=inst)
        op.add_input("iter", tos)
        # Store the indvar into an internal variable
        self.store("indvar", op.add_output("indvar"))

    def _binaryop(self, opname: str, inst: dis.Instruction):
        rhs = self.pop()
        lhs = self.pop()
        op = Op(opname=opname, bc_inst=inst)
        op.add_input("env", self.effect)
        op.add_input("lhs", lhs)
        op.add_input("rhs", rhs)
        self.replace_effect(op.add_output("env", is_effect=True))
        self.push(op.add_output("out"))

    def op_BINARY_OP(self, inst: dis.Instruction):
        self._binaryop("binaryop", inst)

    def op_COMPARE_OP(self, inst: dis.Instruction):
        self._binaryop("compareop", inst)

    def op_IS_OP(self, inst: dis.Instruction):
        self._binaryop("is_op", inst)

    def _unaryop(self, opname: str, inst: dis.Instruction):
        op = Op(opname=opname, bc_inst=inst)
        op.add_input("val", self.pop())
        self.push(op.add_output("out"))

    def op_UNARY_NOT(self, inst: dis.Instruction):
        self._unaryop("not", inst)

    def op_BINARY_SUBSCR(self, inst: dis.Instruction):
        index = self.pop()
        target = self.pop()
        op = Op(opname="binary_subscr", bc_inst=inst)
        op.add_input("env", self.effect)
        op.add_input("index", index)
        op.add_input("target", target)
        self.replace_effect(op.add_output("env", is_effect=True))
        self.push(op.add_output("out"))

    def op_STORE_SUBSCR(self, inst: dis.Instruction):
        index = self.pop()
        target = self.pop()
        value = self.pop()
        op = Op(opname="store_subscr", bc_inst=inst)
        op.add_input("env", self.effect)
        op.add_input("index", index)
        op.add_input("target", target)
        op.add_input("value", value)
        self.replace_effect(op.add_output("env", is_effect=True))

    def op_BUILD_TUPLE(self, inst: dis.Instruction):
        count = inst.arg
        assert isinstance(count, int)
        items = list(reversed([self.pop() for _ in range(count)]))
        op = Op(opname="build_tuple", bc_inst=inst)
        for i, it in enumerate(items):
            op.add_input(str(i), it)
        self.push(op.add_output("out"))

    def op_BUILD_SLICE(self, inst: dis.Instruction):
        argc = inst.arg
        if argc == 2:
            tos = self.pop()
            tos1 = self.pop()
            start = tos1
            stop = tos
            step = None
        elif argc == 3:
            tos = self.pop()
            tos1 = self.pop()
            tos2 = self.pop()
            start = tos2
            stop = tos1
            step = tos
        else:
            raise Exception("unreachable")

        op = Op(opname="build_slice", bc_inst=inst)
        op.add_input("start", start)
        op.add_input("stop", stop)
        if step is not None:
            op.add_input("step", step)
        self.push(op.add_output("out"))

    def op_RETURN_VALUE(self, inst: dis.Instruction):
        tos = self.pop()
        op = Op(opname="ret", bc_inst=inst)
        op.add_input("env", self.effect)
        op.add_input("retval", tos)
        self.replace_effect(op.add_output("env", is_effect=True))

    def op_RAISE_VARARGS(self, inst: dis.Instruction):
        if inst.arg == 0:
            exc = None
            # # No re-raising within a try-except block.
            # # But we allow bare reraise.
            # if state.has_active_try():
            #     raise UnsupportedError(
            #         "The re-raising of an exception is not yet supported.",
            #         loc=self.get_debug_loc(inst.lineno),
            #     )
            raise NotImplementedError
        elif inst.arg == 1:
            exc = self.pop()
        else:
            raise ValueError("Multiple argument raise is not supported.")
        op = Op(opname="raise_varargs", bc_inst=inst)
        op.add_input("env", self.effect)
        op.add_input("exc", exc)
        self.replace_effect(op.add_output("env", is_effect=True))

    def op_JUMP_FORWARD(self, inst: dis.Instruction):
        pass  # no-op

    def op_JUMP_BACKWARD(self, inst: dis.Instruction):
        pass  # no-op

    def _POP_JUMP_X_IF_Y(self, inst: dis.Instruction, *, opname: str):
        tos = self.pop()
        op = Op(opname, bc_inst=inst)
        op.add_input("env", self.effect)
        op.add_input("pred", tos)
        self.replace_effect(op.add_output("env", is_effect=True))

    def op_POP_JUMP_FORWARD_IF_TRUE(self, inst: dis.Instruction):
        self._POP_JUMP_X_IF_Y(inst, opname="jump.if_true")

    def op_POP_JUMP_FORWARD_IF_FALSE(self, inst: dis.Instruction):
        self._POP_JUMP_X_IF_Y(inst, opname="jump.if_false")

    def op_POP_JUMP_BACKWARD_IF_TRUE(self, inst: dis.Instruction):
        self._POP_JUMP_X_IF_Y(inst, opname="jump.if_true")

    def op_POP_JUMP_BACKWARD_IF_FALSE(self, inst: dis.Instruction):
        self._POP_JUMP_X_IF_Y(inst, opname="jump.if_false")

    def op_POP_JUMP_FORWARD_IF_NONE(self, inst: dis.Instruction):
        self._POP_JUMP_X_IF_Y(inst, opname="jump.if_none")

    def op_POP_JUMP_FORWARD_IF_NOT_NONE(self, inst: dis.Instruction):
        self._POP_JUMP_X_IF_Y(inst, opname="jump.if_not_none")

    def _JUMP_IF_X_OR_POP(self, inst: dis.Instruction, *, opname):
        tos = self.top()
        op = Op(opname, bc_inst=inst)
        op.add_input("env", self.effect)
        op.add_input("pred", tos)
        self.replace_effect(op.add_output("env", is_effect=True))

    def op_JUMP_IF_TRUE_OR_POP(self, inst: dis.Instruction):
        self._JUMP_IF_X_OR_POP(inst, opname="jump.if_true")

    def op_JUMP_IF_FALSE_OR_POP(self, inst: dis.Instruction):
        self._JUMP_IF_X_OR_POP(inst, opname="jump.if_false")



# NOTE: This test is sensitive to line numbers as it checks breakpoints
from numba import njit
from numba.tests.gdb_support import GdbMIDriver
from numba.tests.support import TestCase, needs_subprocess
import unittest


@needs_subprocess
class Test(TestCase):

    def test(self):

        @njit(debug=True)
        def foo(x, y):
            c = x + y  # break-here
            return c

        @njit(debug=True)
        def call_foo(a):
            acc = 0
            for i in range(10):
                acc += foo(i, a)
            return acc

        call_foo(10)

        driver = GdbMIDriver(__file__)
        driver.set_breakpoint(line=15, condition='x == 4')
        driver.run()
        driver.check_hit_breakpoint(1)
        driver.stack_list_arguments(1)
        expect = ('[frame={level="0",args=[{name="x",value="4"},'
                  '{name="y",value="10"}]}]')
        driver.assert_output(expect)
        driver.set_breakpoint(line=22, condition='i == 8')
        driver.cont()
        driver.check_hit_breakpoint(2)
        driver.stack_list_variables(1)
        # i should be 8
        driver.assert_output('{name="i",value="8"}')
        driver.quit()


if __name__ == '__main__':
    unittest.main()



"""gdb printing extension for Numba types.
"""
import re

try:
    import gdb.printing
    import gdb
except ImportError:
    raise ImportError("GDB python support is not available.")


class NumbaArrayPrinter:

    def __init__(self, val):
        self.val = val

    def to_string(self):
        try:
            import numpy as np
            HAVE_NUMPY = True
        except ImportError:
            HAVE_NUMPY = False

        try:
            NULL = 0x0

            # Raw data references, these need unpacking/interpreting.

            # Member "data" is...
            # DW_TAG_member of DIDerivedType, tag of DW_TAG_pointer_type
            # encoding e.g. DW_ATE_float
            data = self.val["data"]

            # Member "itemsize" is...
            # DW_TAG_member of DIBasicType encoding DW_ATE_signed
            itemsize = self.val["itemsize"]

            # Members "shape" and "strides" are...
            # DW_TAG_member of DIDerivedType, the type is a DICompositeType
            # (it's a Numba UniTuple) with tag: DW_TAG_array_type, i.e. it's an
            # array repr, it has a basetype of e.g. DW_ATE_unsigned and also
            # "elements" which are referenced with a DISubrange(count: <const>)
            # to say how many elements are in the array.
            rshp = self.val["shape"]
            rstrides = self.val["strides"]

            # bool on whether the data is aligned.
            is_aligned = False

            # type information decode, simple type:
            ty_str = str(self.val.type)
            if HAVE_NUMPY and ('aligned' in ty_str or 'Record' in ty_str):
                ty_str = ty_str.replace('unaligned ','').strip()
                matcher = re.compile(r"array\((Record.*), (.*), (.*)\)\ \(.*")
                # NOTE: need to deal with "Alignment" else dtype size is wrong
                arr_info = [x.strip() for x in matcher.match(ty_str).groups()]
                dtype_str, ndim_str, order_str = arr_info
                rstr = r'Record\((.*\[.*\]);([0-9]+);(True|False)'
                rstr_match = re.match(rstr, dtype_str)
                # balign is unused, it's the alignment
                fields, balign, is_aligned_str = rstr_match.groups()
                is_aligned = is_aligned_str == 'True'
                field_dts = fields.split(',')
                struct_entries = []
                for f in field_dts:
                    splitted = f.split('[')
                    name = splitted[0]
                    dt_part = splitted[1:]
                    if len(dt_part) > 1:
                        raise TypeError('Unsupported sub-type: %s' % f)
                    else:
                        dt_part = dt_part[0]
                        if "nestedarray" in dt_part:
                            raise TypeError('Unsupported sub-type: %s' % f)
                        dt_as_str = dt_part.split(';')[0].split('=')[1]
                        dtype = np.dtype(dt_as_str)
                    struct_entries.append((name, dtype))
                    # The dtype is actually a record of some sort
                dtype_str = struct_entries
            else:  # simple type
                matcher = re.compile(r"array\((.*),(.*),(.*)\)\ \(.*")
                arr_info = [x.strip() for x in matcher.match(ty_str).groups()]
                dtype_str, ndim_str, order_str = arr_info
                # fix up unichr dtype
                if 'unichr x ' in dtype_str:
                    dtype_str = dtype_str[1:-1].replace('unichr x ', '<U')

            def dwarr2inttuple(dwarr):
                # Converts a gdb handle to a dwarf array to a tuple of ints
                fields = dwarr.type.fields()
                lo, hi = fields[0].type.range()
                return tuple([int(dwarr[x]) for x in range(lo, hi + 1)])

            # shape/strides extraction
            shape = dwarr2inttuple(rshp)
            strides = dwarr2inttuple(rstrides)

            # if data is not NULL
            if data != NULL:
                if HAVE_NUMPY:
                    # The data extent in bytes is:
                    # sum(shape * strides)
                    # get the data, then wire to as_strided
                    shp_arr = np.array([max(0, x - 1) for x in shape])
                    strd_arr = np.array(strides)
                    extent = np.sum(shp_arr * strd_arr)
                    extent += int(itemsize)
                    dtype_clazz = np.dtype(dtype_str, align=is_aligned)
                    dtype = dtype_clazz
                    this_proc = gdb.selected_inferior()
                    mem = this_proc.read_memory(int(data), extent)
                    arr_data = np.frombuffer(mem, dtype=dtype)
                    new_arr = np.lib.stride_tricks.as_strided(arr_data,
                                                              shape=shape,
                                                              strides=strides,)
                    return '\n' + str(new_arr)
                # Catch all for no NumPy
                return "array([...], dtype=%s, shape=%s)" % (dtype_str, shape)
            else:
                # Not yet initialized or NULLed out data
                buf = list(["NULL/Uninitialized"])
                return "array([" + ', '.join(buf) + "]" + ")"
        except Exception as e:
            return 'array[Exception: Failed to parse. %s]' % e


class NumbaComplexPrinter:

    def __init__(self, val):
        self.val = val

    def to_string(self):
        return "%s+%sj" % (self.val['real'], self.val['imag'])


class NumbaTuplePrinter:

    def __init__(self, val):
        self.val = val

    def to_string(self):
        buf = []
        fields = self.val.type.fields()
        for f in fields:
            buf.append(str(self.val[f.name]))
        return "(%s)" % ', '.join(buf)


class NumbaUniTuplePrinter:

    def __init__(self, val):
        self.val = val

    def to_string(self):
        # unituples are arrays
        fields = self.val.type.fields()
        lo, hi = fields[0].type.range()
        buf = []
        for i in range(lo, hi + 1):
            buf.append(str(self.val[i]))
        return "(%s)" % ', '.join(buf)


class NumbaUnicodeTypePrinter:

    def __init__(self, val):
        self.val = val

    def to_string(self):
        NULL = 0x0
        data = self.val["data"]
        nitems = self.val["length"]
        kind = self.val["kind"]
        if data != NULL:
            # This needs sorting out, encoding is wrong
            this_proc = gdb.selected_inferior()
            mem = this_proc.read_memory(int(data), nitems * kind)
            if isinstance(mem, memoryview):
                buf = bytes(mem).decode()
            else:
                buf = mem.decode('utf-8')
        else:
            buf = str(data)
        return "'%s'" % buf


def _create_printers():
    printer = gdb.printing.RegexpCollectionPrettyPrinter("Numba")
    printer.add_printer('Numba unaligned array printer', '^unaligned array\\(',
                        NumbaArrayPrinter)
    printer.add_printer('Numba array printer', '^array\\(', NumbaArrayPrinter)
    printer.add_printer('Numba complex printer', '^complex[0-9]+\\ ',
                        NumbaComplexPrinter)
    printer.add_printer('Numba Tuple printer', '^Tuple\\(',
                        NumbaTuplePrinter)
    printer.add_printer('Numba UniTuple printer', '^UniTuple\\(',
                        NumbaUniTuplePrinter)
    printer.add_printer('Numba unicode_type printer', '^unicode_type\\s+\\(',
                        NumbaUnicodeTypePrinter)
    return printer


# register the Numba pretty printers for the current object
gdb.printing.register_pretty_printer(gdb.current_objfile(), _create_printers())



"""
Testing numba implementation of the numba dictionary.

The tests here only check that the numba typing and codegen are working
correctly.  Detailed testing of the underlying dictionary operations is done
in test_dictimpl.py.
"""

import sys
import warnings

import numpy as np

from numba import njit, literally
from numba import int32, int64, float32, float64
from numba import typeof
from numba.typed import Dict, dictobject, List
from numba.typed.typedobjectutils import _sentry_safe_cast
from numba.core.errors import TypingError
from numba.core import types
from numba.tests.support import (TestCase, MemoryLeakMixin, unittest,
                                 override_config, forbid_codegen)
from numba.experimental import jitclass
from numba.extending import overload


class TestDictObject(MemoryLeakMixin, TestCase):
    def test_dict_bool(self):
        """
        Exercise bool(dict)
        """
        @njit
        def foo(n):
            d = dictobject.new_dict(int32, float32)
            for i in range(n):
                d[i] = i + 1
            return bool(d)

        # Insert nothing
        self.assertEqual(foo(n=0), False)
        # Insert 1 entry
        self.assertEqual(foo(n=1), True)
        # Insert 2 entries
        self.assertEqual(foo(n=2), True)
        # Insert 100 entries
        self.assertEqual(foo(n=100), True)

    def test_dict_create(self):
        """
        Exercise dictionary creation, insertion and len
        """
        @njit
        def foo(n):
            d = dictobject.new_dict(int32, float32)
            for i in range(n):
                d[i] = i + 1
            return len(d)

        # Insert nothing
        self.assertEqual(foo(n=0), 0)
        # Insert 1 entry
        self.assertEqual(foo(n=1), 1)
        # Insert 2 entries
        self.assertEqual(foo(n=2), 2)
        # Insert 100 entries
        self.assertEqual(foo(n=100), 100)

    def test_dict_get(self):
        """
        Exercise dictionary creation, insertion and get
        """
        @njit
        def foo(n, targets):
            d = dictobject.new_dict(int32, float64)
            # insertion loop
            for i in range(n):
                d[i] = i
            # retrieval loop
            output = []
            for t in targets:
                output.append(d.get(t))
            return output

        self.assertEqual(foo(5, [0, 1, 9]), [0, 1, None])
        self.assertEqual(foo(10, [0, 1, 9]), [0, 1, 9])
        self.assertEqual(foo(10, [-1, 9, 1]), [None, 9, 1])

    def test_dict_get_with_default(self):
        """
        Exercise dict.get(k, d) where d is set
        """
        @njit
        def foo(n, target, default):
            d = dictobject.new_dict(int32, float64)
            # insertion loop
            for i in range(n):
                d[i] = i
            # retrieval loop
            return d.get(target, default)

        self.assertEqual(foo(5, 3, -1), 3)
        self.assertEqual(foo(5, 5, -1), -1)

    def test_dict_getitem(self):
        """
        Exercise dictionary __getitem__
        """
        @njit
        def foo(keys, vals, target):
            d = dictobject.new_dict(int32, float64)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v

            # lookup
            return d[target]

        keys = [1, 2, 3]
        vals = [0.1, 0.2, 0.3]
        self.assertEqual(foo(keys, vals, 1), 0.1)
        self.assertEqual(foo(keys, vals, 2), 0.2)
        self.assertEqual(foo(keys, vals, 3), 0.3)
        # check no leak so far
        self.assert_no_memory_leak()
        # disable leak check for exception test
        self.disable_leak_check()
        with self.assertRaises(KeyError):
            foo(keys, vals, 0)
        with self.assertRaises(KeyError):
            foo(keys, vals, 4)

    def test_dict_popitem(self):
        """
        Exercise dictionary .popitem
        """
        @njit
        def foo(keys, vals):
            d = dictobject.new_dict(int32, float64)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v

            # popitem
            return d.popitem()

        keys = [1, 2, 3]
        vals = [0.1, 0.2, 0.3]
        for i in range(1, len(keys)):
            self.assertEqual(
                foo(keys[:i], vals[:i]),
                (keys[i - 1], vals[i - 1]),
            )

    def test_dict_popitem_many(self):
        """
        Exercise dictionary .popitem
        """

        @njit
        def core(d, npop):
            # popitem
            keysum, valsum = 0, 0
            for _ in range(npop):
                k, v = d.popitem()
                keysum += k
                valsum -= v
            return keysum, valsum

        @njit
        def foo(keys, vals, npop):
            d = dictobject.new_dict(int32, int32)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v

            return core(d, npop)

        keys = [1, 2, 3]
        vals = [10, 20, 30]

        for i in range(len(keys)):
            self.assertEqual(
                foo(keys, vals, npop=3),
                core.py_func(dict(zip(keys, vals)), npop=3),
            )

        # check no leak so far
        self.assert_no_memory_leak()
        # disable leak check for exception test
        self.disable_leak_check()

        with self.assertRaises(KeyError):
            foo(keys, vals, npop=4)

    def test_dict_pop(self):
        """
        Exercise dictionary .pop
        """
        @njit
        def foo(keys, vals, target):
            d = dictobject.new_dict(int32, float64)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v

            # popitem
            return d.pop(target, None), len(d)

        keys = [1, 2, 3]
        vals = [0.1, 0.2, 0.3]

        self.assertEqual(foo(keys, vals, 1), (0.1, 2))
        self.assertEqual(foo(keys, vals, 2), (0.2, 2))
        self.assertEqual(foo(keys, vals, 3), (0.3, 2))
        self.assertEqual(foo(keys, vals, 0), (None, 3))

        # check no leak so far
        self.assert_no_memory_leak()
        # disable leak check for exception test
        self.disable_leak_check()

        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            # popitem
            return d.pop(0)

        with self.assertRaises(KeyError):
            foo()

    def test_dict_pop_many(self):
        """
        Exercise dictionary .pop
        """

        @njit
        def core(d, pops):
            total = 0
            for k in pops:
                total += k + d.pop(k, 0.123) + len(d)
                total *= 2
            return total

        @njit
        def foo(keys, vals, pops):
            d = dictobject.new_dict(int32, float64)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v
            # popitem
            return core(d, pops)

        keys = [1, 2, 3]
        vals = [0.1, 0.2, 0.3]
        pops = [2, 3, 3, 1, 0, 2, 1, 0, -1]

        self.assertEqual(
            foo(keys, vals, pops),
            core.py_func(dict(zip(keys, vals)), pops),
        )

    def test_dict_delitem(self):
        @njit
        def foo(keys, vals, target):
            d = dictobject.new_dict(int32, float64)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v
            del d[target]
            return len(d), d.get(target)

        keys = [1, 2, 3]
        vals = [0.1, 0.2, 0.3]
        self.assertEqual(foo(keys, vals, 1), (2, None))
        self.assertEqual(foo(keys, vals, 2), (2, None))
        self.assertEqual(foo(keys, vals, 3), (2, None))
        # check no leak so far
        self.assert_no_memory_leak()
        # disable leak check for exception test
        self.disable_leak_check()
        with self.assertRaises(KeyError):
            foo(keys, vals, 0)

    def test_dict_clear(self):
        """
        Exercise dict.clear
        """
        @njit
        def foo(keys, vals):
            d = dictobject.new_dict(int32, float64)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v
            b4 = len(d)
            # clear
            d.clear()
            return b4, len(d)

        keys = [1, 2, 3]
        vals = [0.1, 0.2, 0.3]
        self.assertEqual(foo(keys, vals), (3, 0))

    def test_dict_items(self):
        """
        Exercise dict.items
        """
        @njit
        def foo(keys, vals):
            d = dictobject.new_dict(int32, float64)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v
            out = []
            for kv in d.items():
                out.append(kv)
            return out

        keys = [1, 2, 3]
        vals = [0.1, 0.2, 0.3]

        self.assertEqual(
            foo(keys, vals),
            list(zip(keys, vals)),
        )

        # Test .items() on empty dict
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            out = []
            for kv in d.items():
                out.append(kv)
            return out

        self.assertEqual(foo(), [])

    def test_dict_keys(self):
        """
        Exercise dict.keys
        """
        @njit
        def foo(keys, vals):
            d = dictobject.new_dict(int32, float64)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v
            out = []
            for k in d.keys():
                out.append(k)
            return out

        keys = [1, 2, 3]
        vals = [0.1, 0.2, 0.3]

        self.assertEqual(
            foo(keys, vals),
            keys,
        )

    def test_dict_keys_len(self):
        """
        Exercise len(dict.keys())
        """
        @njit
        def foo(keys, vals):
            d = dictobject.new_dict(int32, float64)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v
            return len(d.keys())

        keys = [1, 2, 3]
        vals = [0.1, 0.2, 0.3]

        self.assertEqual(
            foo(keys, vals),
            len(keys),
        )

    def test_dict_values(self):
        """
        Exercise dict.values
        """
        @njit
        def foo(keys, vals):
            d = dictobject.new_dict(int32, float64)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v
            out = []
            for v in d.values():
                out.append(v)
            return out

        keys = [1, 2, 3]
        vals = [0.1, 0.2, 0.3]

        self.assertEqual(
            foo(keys, vals),
            vals,
        )

    def test_dict_values_len(self):
        """
        Exercise len(dict.values())
        """
        @njit
        def foo(keys, vals):
            d = dictobject.new_dict(int32, float64)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v
            return len(d.values())

        keys = [1, 2, 3]
        vals = [0.1, 0.2, 0.3]

        self.assertEqual(
            foo(keys, vals),
            len(vals),
        )

    def test_dict_items_len(self):
        """
        Exercise len(dict.items())
        """
        @njit
        def foo(keys, vals):
            d = dictobject.new_dict(int32, float64)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v
            return len(d.items())

        keys = [1, 2, 3]
        vals = [0.1, 0.2, 0.3]
        self.assertPreciseEqual(
            foo(keys, vals),
            len(vals),
        )

    def test_dict_iter(self):
        """
        Exercise iter(dict)
        """
        @njit
        def foo(keys, vals):
            d = dictobject.new_dict(int32, float64)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v
            out = []
            for k in d:
                out.append(k)
            return out

        keys = [1, 2, 3]
        vals = [0.1, 0.2, 0.3]

        self.assertEqual(
            foo(keys, vals),
            [1, 2, 3]
        )

    def test_dict_contains(self):
        """
        Exercise operator.contains
        """
        @njit
        def foo(keys, vals, checklist):
            d = dictobject.new_dict(int32, float64)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v
            out = []
            for k in checklist:
                out.append(k in d)
            return out

        keys = [1, 2, 3]
        vals = [0.1, 0.2, 0.3]

        self.assertEqual(
            foo(keys, vals, [2, 3, 4, 1, 0]),
            [True, True, False, True, False],
        )

    def test_dict_copy(self):
        """
        Exercise dict.copy
        """
        @njit
        def foo(keys, vals):
            d = dictobject.new_dict(int32, float64)
            # insertion
            for k, v in zip(keys, vals):
                d[k] = v
            return list(d.copy().items())

        keys = list(range(20))
        vals = [x + i / 100 for i, x in enumerate(keys)]
        out = foo(keys, vals)
        self.assertEqual(out, list(zip(keys, vals)))

    def test_dict_setdefault(self):
        """
        Exercise dict.setdefault
        """
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            d.setdefault(1, 1.2)  # used because key is not in
            a = d.get(1)
            d[1] = 2.3
            b = d.get(1)
            d[2] = 3.4
            d.setdefault(2, 4.5)  # not used because key is in
            c = d.get(2)
            return a, b, c

        self.assertEqual(foo(), (1.2, 2.3, 3.4))

    def test_dict_equality(self):
        """
        Exercise dict.__eq__ and .__ne__
        """
        @njit
        def foo(na, nb, fa, fb):
            da = dictobject.new_dict(int32, float64)
            db = dictobject.new_dict(int32, float64)
            for i in range(na):
                da[i] = i * fa
            for i in range(nb):
                db[i] = i * fb
            return da == db, da != db

        # Same keys and values
        self.assertEqual(foo(10, 10, 3, 3), (True, False))
        # Same keys and diff values
        self.assertEqual(foo(10, 10, 3, 3.1), (False, True))
        # LHS has more keys
        self.assertEqual(foo(11, 10, 3, 3), (False, True))
        # RHS has more keys
        self.assertEqual(foo(10, 11, 3, 3), (False, True))

    def test_dict_equality_more(self):
        """
        Exercise dict.__eq__
        """
        @njit
        def foo(ak, av, bk, bv):
            # The key-value types are different in the two dictionaries
            da = dictobject.new_dict(int32, float64)
            db = dictobject.new_dict(int64, float32)
            for i in range(len(ak)):
                da[ak[i]] = av[i]
            for i in range(len(bk)):
                db[bk[i]] = bv[i]
            return da == db

        # Simple equal case
        ak = [1, 2, 3]
        av = [2, 3, 4]
        bk = [1, 2, 3]
        bv = [2, 3, 4]
        self.assertTrue(foo(ak, av, bk, bv))

        # Equal with replacement
        ak = [1, 2, 3]
        av = [2, 3, 4]
        bk = [1, 2, 2, 3]
        bv = [2, 1, 3, 4]
        self.assertTrue(foo(ak, av, bk, bv))

        # Diff values
        ak = [1, 2, 3]
        av = [2, 3, 4]
        bk = [1, 2, 3]
        bv = [2, 1, 4]
        self.assertFalse(foo(ak, av, bk, bv))

        # Diff keys
        ak = [0, 2, 3]
        av = [2, 3, 4]
        bk = [1, 2, 3]
        bv = [2, 3, 4]
        self.assertFalse(foo(ak, av, bk, bv))

    def test_dict_equality_diff_type(self):
        """
        Exercise dict.__eq__
        """
        @njit
        def foo(na, b):
            da = dictobject.new_dict(int32, float64)
            for i in range(na):
                da[i] = i
            return da == b

        # dict != int
        self.assertFalse(foo(10, 1))
        # dict != tuple[int]
        self.assertFalse(foo(10, (1,)))

    def test_dict_to_from_meminfo(self):
        """
        Exercise dictobject.{_as_meminfo, _from_meminfo}
        """
        @njit
        def make_content(nelem):
            for i in range(nelem):
                yield i, i + (i + 1) / 100

        @njit
        def boxer(nelem):
            d = dictobject.new_dict(int32, float64)
            for k, v in make_content(nelem):
                d[k] = v
            return dictobject._as_meminfo(d)

        dcttype = types.DictType(int32, float64)

        @njit
        def unboxer(mi):
            d = dictobject._from_meminfo(mi, dcttype)
            return list(d.items())

        mi = boxer(10)
        self.assertEqual(mi.refcount, 1)

        got = unboxer(mi)
        expected = list(make_content.py_func(10))
        self.assertEqual(got, expected)

    def test_001_cannot_downcast_key(self):
        @njit
        def foo(n):
            d = dictobject.new_dict(int32, float64)
            for i in range(n):
                d[i] = i + 1
            # bad key type
            z = d.get(1j)
            return z

        with self.assertRaises(TypingError) as raises:
            foo(10)
        self.assertIn(
            'cannot safely cast complex128 to int32',
            str(raises.exception),
        )

    def test_002_cannot_downcast_default(self):
        @njit
        def foo(n):
            d = dictobject.new_dict(int32, float64)
            for i in range(n):
                d[i] = i + 1
            # bad default type
            z = d.get(2 * n, 1j)
            return z

        with self.assertRaises(TypingError) as raises:
            foo(10)
        self.assertIn(
            'cannot safely cast complex128 to float64',
            str(raises.exception),
        )

    def test_003_cannot_downcast_key(self):
        @njit
        def foo(n):
            d = dictobject.new_dict(int32, float64)
            for i in range(n):
                d[i] = i + 1
            # bad cast!?
            z = d.get(2.4)
            return z

        # should raise
        with self.assertRaises(TypingError) as raises:
            foo(10)
        self.assertIn(
            'cannot safely cast float64 to int32',
            str(raises.exception),
        )

    def test_004_cannot_downcast_key(self):
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            # should raise TypingError
            d[1j] = 7.

        with self.assertRaises(TypingError) as raises:
            foo()
        self.assertIn(
            'cannot safely cast complex128 to int32',
            str(raises.exception),
        )

    def test_005_cannot_downcast_value(self):
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            # should raise TypingError
            d[1] = 1j

        with self.assertRaises(TypingError) as raises:
            foo()
        self.assertIn(
            'cannot safely cast complex128 to float64',
            str(raises.exception),
        )

    def test_006_cannot_downcast_key(self):
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            # raise TypingError
            d[11.5]

        with self.assertRaises(TypingError) as raises:
            foo()
        self.assertIn(
            'cannot safely cast float64 to int32',
            str(raises.exception),
        )

    @unittest.skipUnless(sys.maxsize > 2 ** 32, "64 bit test only")
    def test_007_collision_checks(self):
        # this checks collisions in real life for 64bit systems
        @njit
        def foo(v1, v2):
            d = dictobject.new_dict(int64, float64)
            c1 = np.uint64(2 ** 61 - 1)
            c2 = np.uint64(0)
            assert hash(c1) == hash(c2)
            d[c1] = v1
            d[c2] = v2
            return (d[c1], d[c2])

        a, b = 10., 20.
        x, y = foo(a, b)
        self.assertEqual(x, a)
        self.assertEqual(y, b)

    def test_008_lifo_popitem(self):
        # check that (keys, vals) are LIFO .popitem()
        @njit
        def foo(n):
            d = dictobject.new_dict(int32, float64)
            for i in range(n):
                d[i] = i + 1
            keys = []
            vals = []
            for i in range(n):
                tmp = d.popitem()
                keys.append(tmp[0])
                vals.append(tmp[1])
            return keys, vals

        z = 10
        gk, gv = foo(z)

        self.assertEqual(gk, [x for x in reversed(range(z))])
        self.assertEqual(gv, [x + 1 for x in reversed(range(z))])

    def test_010_cannot_downcast_default(self):
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            d[0] = 6.
            d[1] = 7.
            # pop'd default must have same type as value
            d.pop(11, 12j)

        with self.assertRaises(TypingError) as raises:
            foo()
        self.assertIn(
            "cannot safely cast complex128 to float64",
            str(raises.exception),
        )

    def test_011_cannot_downcast_key(self):
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            d[0] = 6.
            d[1] = 7.
            # pop'd key must have same type as key
            d.pop(11j)

        with self.assertRaises(TypingError) as raises:
            foo()
        self.assertIn(
            "cannot safely cast complex128 to int32",
            str(raises.exception),
        )

    def test_012_cannot_downcast_key(self):
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            d[0] = 6.
            # invalid key type
            return 1j in d

        with self.assertRaises(TypingError) as raises:
            foo()
        self.assertIn(
            "cannot safely cast complex128 to int32",
            str(raises.exception),
        )

    def test_013_contains_empty_dict(self):
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            # contains on empty dict
            return 1 in d

        self.assertFalse(foo())

    def test_014_not_contains_empty_dict(self):
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            # not contains empty dict
            return 1 not in d

        self.assertTrue(foo())

    def test_015_dict_clear(self):
        @njit
        def foo(n):
            d = dictobject.new_dict(int32, float64)
            for i in range(n):
                d[i] = i + 1
            x = len(d)
            d.clear()
            y = len(d)
            return x, y

        m = 10
        self.assertEqual(foo(m), (m, 0))

    def test_016_cannot_downcast_key(self):
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            # key is wrong type
            d.setdefault(1j, 12.)

        with self.assertRaises(TypingError) as raises:
            foo()
        self.assertIn(
            "cannot safely cast complex128 to int32",
            str(raises.exception),
        )

    def test_017_cannot_downcast_default(self):
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            # default value is wrong type
            d.setdefault(1, 12.j)

        with self.assertRaises(TypingError) as raises:
            foo()
        self.assertIn(
            "cannot safely cast complex128 to float64",
            str(raises.exception),
        )

    def test_018_keys_iter_are_views(self):
        # this is broken somewhere in llvmlite, intent of test is to check if
        # keys behaves like a view or not
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            d[11] = 12.
            k1 = d.keys()
            d[22] = 9.
            k2 = d.keys()
            rk1 = [x for x in k1]
            rk2 = [x for x in k2]
            return rk1, rk2

        a, b = foo()
        self.assertEqual(a, b)
        self.assertEqual(a, [11, 22])

    # Not implemented yet
    @unittest.expectedFailure
    def test_019(self):
        # should keys/vals be set-like?
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            d[11] = 12.
            d[22] = 9.
            k2 = d.keys() & {12, }
            return k2

        print(foo())

    def test_020_string_key(self):
        @njit
        def foo():
            d = dictobject.new_dict(types.unicode_type, float64)
            d['a'] = 1.
            d['b'] = 2.
            d['c'] = 3.
            d['d'] = 4.
            out = []
            for x in d.items():
                out.append(x)
            return out, d['a']

        items, da = foo()
        self.assertEqual(items, [('a', 1.), ('b', 2.), ('c', 3.), ('d', 4)])
        self.assertEqual(da, 1.)

    def test_021_long_str_key(self):
        @njit
        def foo():
            d = dictobject.new_dict(types.unicode_type, float64)
            tmp = []
            for i in range(10000):
                tmp.append('a')
            s = ''.join(tmp)
            d[s] = 1.
            out = list(d.items())
            return out
        self.assertEqual(foo(), [('a' * 10000, 1)])

    def test_022_references_juggle(self):
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)
            e = d
            d[1] = 12.
            e[2] = 14.
            e = dictobject.new_dict(int32, float64)
            e[1] = 100.
            e[2] = 1000.
            f = d
            d = e

            k1 = [x for x in d.items()]
            k2 = [x for x in e.items()]
            k3 = [x for x in f.items()]

            return k1, k2, k3

        k1, k2, k3 = foo()
        self.assertEqual(k1, [(1, 100.0), (2, 1000.0)])
        self.assertEqual(k2, [(1, 100.0), (2, 1000.0)])
        self.assertEqual(k3, [(1, 12), (2, 14)])

    def test_023_closure(self):
        @njit
        def foo():
            d = dictobject.new_dict(int32, float64)

            def bar():
                d[1] = 12.
                d[2] = 14.
            bar()
            return [x for x in d.keys()]

        self.assertEqual(foo(), [1, 2])

    def test_024_unicode_getitem_keys(self):
        # See issue #6135
        @njit
        def foo():
            s = 'a\u1234'
            d = {s[0] : 1}
            return d['a']

        self.assertEqual(foo(), foo.py_func())

        @njit
        def foo():
            s = 'abc\u1234'
            d = {s[:1] : 1}
            return d['a']

        self.assertEqual(foo(), foo.py_func())

    def test_issue6570_alignment_padding(self):
        # Create a key type that is 12-bytes long on a 8-byte aligned system
        # so that the a 4-byte padding is needed.
        # If the 4-byte padding is not zero-filled, it will have garbage data
        # that affects key matching in the lookup.
        keyty = types.Tuple([types.uint64, types.float32])

        @njit
        def foo():
            d = dictobject.new_dict(keyty, float64)
            t1 = np.array([3], dtype=np.uint64)
            t2 = np.array([5.67], dtype=np.float32)
            v1 = np.array([10.23], dtype=np.float32)
            d[(t1[0], t2[0])] = v1[0]
            return (t1[0], t2[0]) in d

        self.assertTrue(foo())

    def test_dict_update(self):
        """
        Tests dict.update works with various dictionaries.
        """
        n = 10

        def f1(n):
            """
            Test update with a regular dictionary.
            """
            d1 = {i: i + 1 for i in range(n)}
            d2 = {3 * i: i for i in range(n)}
            d1.update(d2)
            return d1

        py_func = f1
        cfunc = njit()(f1)
        a = py_func(n)
        b = cfunc(n)
        self.assertEqual(a, b)

        def f2(n):
            """
            Test update where one of the dictionaries
            is created as a Python literal.
            """
            d1 = {
                1: 2,
                3: 4,
                5: 6
            }
            d2 = {3 * i: i for i in range(n)}
            d1.update(d2)
            return d1

        py_func = f2
        cfunc = njit()(f2)
        a = py_func(n)
        b = cfunc(n)
        self.assertEqual(a, b)


class TestDictTypeCasting(TestCase):
    def check_good(self, fromty, toty):
        _sentry_safe_cast(fromty, toty)

    def check_bad(self, fromty, toty):
        with self.assertRaises(TypingError) as raises:
            _sentry_safe_cast(fromty, toty)
        self.assertIn(
            'cannot safely cast {fromty} to {toty}'.format(**locals()),
            str(raises.exception),
        )

    def test_cast_int_to(self):
        self.check_good(types.int32, types.float32)
        self.check_good(types.int32, types.float64)
        self.check_good(types.int32, types.complex128)
        self.check_good(types.int64, types.complex128)
        self.check_bad(types.int32, types.complex64)
        self.check_good(types.int8, types.complex64)

    def test_cast_float_to(self):
        self.check_good(types.float32, types.float64)
        self.check_good(types.float32, types.complex64)
        self.check_good(types.float64, types.complex128)

    def test_cast_bool_to(self):
        self.check_good(types.boolean, types.int32)
        self.check_good(types.boolean, types.float64)
        self.check_good(types.boolean, types.complex128)


class TestTypedDict(MemoryLeakMixin, TestCase):
    def test_basic(self):
        d = Dict.empty(int32, float32)
        # len
        self.assertEqual(len(d), 0)
        # setitems
        d[1] = 1
        d[2] = 2.3
        d[3] = 3.4
        self.assertEqual(len(d), 3)
        # keys
        self.assertEqual(list(d.keys()), [1, 2, 3])
        # values
        for x, y in zip(list(d.values()), [1, 2.3, 3.4]):
            self.assertAlmostEqual(x, y, places=4)
        # getitem
        self.assertAlmostEqual(d[1], 1)
        self.assertAlmostEqual(d[2], 2.3, places=4)
        self.assertAlmostEqual(d[3], 3.4, places=4)
        # deltiem
        del d[2]
        self.assertEqual(len(d), 2)
        # get
        self.assertIsNone(d.get(2))
        # setdefault
        d.setdefault(2, 100)
        d.setdefault(3, 200)
        self.assertEqual(d[2], 100)
        self.assertAlmostEqual(d[3], 3.4, places=4)
        # update
        d.update({4: 5, 5: 6})
        self.assertAlmostEqual(d[4], 5)
        self.assertAlmostEqual(d[5], 6)
        # contains
        self.assertTrue(4 in d)
        # items
        pyd = dict(d.items())
        self.assertEqual(len(pyd), len(d))
        # pop
        self.assertAlmostEqual(d.pop(4), 5)
        # popitem
        nelem = len(d)
        k, v = d.popitem()
        self.assertEqual(len(d), nelem - 1)
        self.assertTrue(k not in d)
        # __eq__ & copy
        copied = d.copy()
        self.assertEqual(copied, d)
        self.assertEqual(list(copied.items()), list(d.items()))

    def test_copy_from_dict(self):
        expect = {k: float(v) for k, v in zip(range(10), range(10, 20))}
        nbd = Dict.empty(int32, float64)
        for k, v in expect.items():
            nbd[k] = v
        got = dict(nbd)
        self.assertEqual(got, expect)

    def test_compiled(self):
        @njit
        def producer():
            d = Dict.empty(int32, float64)
            d[1] = 1.23
            return d

        @njit
        def consumer(d):
            return d[1]

        d = producer()
        val = consumer(d)
        self.assertEqual(val, 1.23)

    def test_gh7908(self):
        d = Dict.empty(
            key_type=types.Tuple([types.uint32,
                                  types.uint32]),
            value_type=int64)

        d[(1, 1)] = 12345
        self.assertEqual(d[(1, 1)], d.get((1, 1)))

    def check_stringify(self, strfn, prefix=False):
        nbd = Dict.empty(int32, int32)
        d = {}
        nbd[1] = 2
        d[1] = 2
        checker = self.assertIn if prefix else self.assertEqual
        checker(strfn(d), strfn(nbd))
        nbd[2] = 3
        d[2] = 3
        checker(strfn(d), strfn(nbd))
        for i in range(10, 20):
            nbd[i] = i + 1
            d[i] = i + 1
        checker(strfn(d), strfn(nbd))
        if prefix:
            self.assertTrue(strfn(nbd).startswith('DictType'))

    def test_repr(self):
        self.check_stringify(repr, prefix=True)

    def test_str(self):
        self.check_stringify(str)


class DictIterableCtor:

    def test_iterable_type_constructor(self):
        # https://docs.python.org/3/library/stdtypes.html#dict
        @njit
        def func1(a, b):
            d = Dict(zip(a, b))
            return d

        @njit
        def func2(a_, b):
            a = range(3)
            return Dict(zip(a, b))

        @njit
        def func3(a_, b):
            a = [0, 1, 2]
            return Dict(zip(a, b))

        @njit
        def func4(a, b):
            c = zip(a, b)
            return Dict(zip(a, zip(c, a)))

        @njit
        def func5(a, b):
            return Dict(zip(zip(a, b), b))

        @njit
        def func6(items):
            return Dict(items)

        @njit
        def func7(k, v):
            return Dict({k: v})  # mapping - not supported

        @njit
        def func8(k, v):
            d = Dict()
            d[k] = v
            return d

        def _get_dict(py_dict):
            d = Dict()
            for k, v in py_dict.items():
                d[k] = v
            return d

        vals = (
            (func1, [(0, 1, 2), 'abc'], _get_dict({0: 'a', 1: 'b', 2: 'c'})),
            (func2, [(0, 1, 2), 'abc'], _get_dict({0: 'a', 1: 'b', 2: 'c'})),
            (func3, [(0, 1, 2), 'abc'], _get_dict({0: 'a', 1: 'b', 2: 'c'})),
            (func4, [(0, 1, 2), 'abc'], _get_dict(
                {0: ((0, 'a'), 0), 1: ((1, 'b'), 1), 2: ((2, 'c'), 2)})),
            (func5, [(0, 1, 2), 'abc'], _get_dict(
                {(0, 'a'): 'a', (1, 'b'): 'b', (2, 'c'): 'c'})),
            # (func6, [(),], Dict({})),
            (func6, [((1, 'a'), (3, 'b')),], _get_dict({1: 'a', 3: 'b'})),
            (func1, ['key', _get_dict({1: 'abc'})], _get_dict({'k': 1})),
            (func8, ['key', _get_dict({1: 'abc'})], _get_dict(
                {'key': _get_dict({1: 'abc'})})),
            (func8, ['key', List([1, 2, 3])], _get_dict(
                {'key': List([1, 2, 3])})),
        )

        for func, args, expected in vals:
            if self.jit_enabled:
                got = func(*args)
            else:
                got = func.py_func(*args)
            self.assertPreciseEqual(expected, got)


class TestDictIterableCtorJit(TestCase, DictIterableCtor):

    def setUp(self):
        self.jit_enabled = True

    def test_exception_no_iterable_arg(self):
        @njit
        def ctor():
            return Dict(3)

        msg = ".*No implementation of function.*"
        with self.assertRaisesRegex(TypingError, msg):
            ctor()

    def test_exception_dict_mapping(self):
        @njit
        def ctor():
            return Dict({1: 2, 3: 4})

        msg = ".*No implementation of function.*"
        with self.assertRaisesRegex(TypingError, msg):
            ctor()

    def test_exception_setitem(self):
        @njit
        def ctor():
            return Dict(((1, 'a'), (2, 'b', 3)))

        msg = ".*No implementation of function.*"
        with self.assertRaisesRegex(TypingError, msg):
            ctor()


class TestDictIterableCtorNoJit(TestCase, DictIterableCtor):

    def setUp(self):
        self.jit_enabled = False

    def test_exception_nargs(self):
        msg = 'Dict expect at most 1 argument, got 2'
        with self.assertRaisesRegex(TypingError, msg):
            Dict(1, 2)

    def test_exception_mapping_ctor(self):
        msg = r'.*dict\(mapping\) is not supported.*'  # noqa: W605
        with self.assertRaisesRegex(TypingError, msg):
            Dict({1: 2})

    def test_exception_non_iterable_arg(self):
        msg = '.*object is not iterable.*'
        with self.assertRaisesRegex(TypingError, msg):
            Dict(3)

    def test_exception_setitem(self):
        msg = ".*dictionary update sequence element #1 has length 3.*"
        with self.assertRaisesRegex(ValueError, msg):
            Dict(((1, 'a'), (2, 'b', 3)))


class TestDictRefctTypes(MemoryLeakMixin, TestCase):

    def test_str_key(self):
        @njit
        def foo():
            d = Dict.empty(
                key_type=types.unicode_type,
                value_type=types.int32,
            )
            d["123"] = 123
            d["321"] = 321
            return d

        d = foo()
        self.assertEqual(d['123'], 123)
        self.assertEqual(d['321'], 321)
        expect = {'123': 123, '321': 321}
        self.assertEqual(dict(d), expect)
        # Test insert replacement
        d['123'] = 231
        expect['123'] = 231
        self.assertEqual(d['123'], 231)
        self.assertEqual(dict(d), expect)
        # Test dictionary growth
        nelem = 100
        for i in range(nelem):
            d[str(i)] = i
            expect[str(i)] = i
        for i in range(nelem):
            self.assertEqual(d[str(i)], i)
        self.assertEqual(dict(d), expect)

    def test_str_val(self):
        @njit
        def foo():
            d = Dict.empty(
                key_type=types.int32,
                value_type=types.unicode_type,
            )
            d[123] = "123"
            d[321] = "321"
            return d

        d = foo()
        self.assertEqual(d[123], '123')
        self.assertEqual(d[321], '321')
        expect = {123: '123', 321: '321'}
        self.assertEqual(dict(d), expect)
        # Test insert replacement
        d[123] = "231"
        expect[123] = "231"
        self.assertEqual(dict(d), expect)
        # Test dictionary growth
        nelem = 1
        for i in range(nelem):
            d[i] = str(i)
            expect[i] = str(i)
        for i in range(nelem):
            self.assertEqual(d[i], str(i))
        self.assertEqual(dict(d), expect)

    def test_str_key_array_value(self):
        np.random.seed(123)
        d = Dict.empty(
            key_type=types.unicode_type,
            value_type=types.float64[:],
        )
        expect = []
        expect.append(np.random.random(10))
        d['mass'] = expect[-1]
        expect.append(np.random.random(20))
        d['velocity'] = expect[-1]
        for i in range(100):
            expect.append(np.random.random(i))
            d[str(i)] = expect[-1]
        self.assertEqual(len(d), len(expect))
        self.assertPreciseEqual(d['mass'], expect[0])
        self.assertPreciseEqual(d['velocity'], expect[1])
        # Ordering is kept
        for got, exp in zip(d.values(), expect):
            self.assertPreciseEqual(got, exp)

        # Try deleting
        self.assertTrue('mass' in d)
        self.assertTrue('velocity' in d)
        del d['mass']
        self.assertFalse('mass' in d)
        del d['velocity']
        self.assertFalse('velocity' in d)
        del expect[0:2]

        for i in range(90):
            k, v = d.popitem()
            w = expect.pop()
            self.assertPreciseEqual(v, w)

        # Trigger a resize
        expect.append(np.random.random(10))
        d["last"] = expect[-1]

        # Ordering is kept
        for got, exp in zip(d.values(), expect):
            self.assertPreciseEqual(got, exp)

    def test_dict_of_dict_int_keyval(self):
        def inner_numba_dict():
            d = Dict.empty(
                key_type=types.intp,
                value_type=types.intp,
            )
            return d

        d = Dict.empty(
            key_type=types.intp,
            value_type=types.DictType(types.intp, types.intp),
        )

        def usecase(d, make_inner_dict):
            for i in range(100):
                mid = make_inner_dict()
                for j in range(i + 1):
                    mid[j] = j * 10000
                d[i] = mid
            return d

        got = usecase(d, inner_numba_dict)
        expect = usecase({}, dict)

        self.assertIsInstance(expect, dict)

        self.assertEqual(dict(got), expect)

        # Delete items
        for where in [12, 3, 6, 8, 10]:
            del got[where]
            del expect[where]
            self.assertEqual(dict(got), expect)

    def test_dict_of_dict_npm(self):
        inner_dict_ty = types.DictType(types.intp, types.intp)

        @njit
        def inner_numba_dict():
            d = Dict.empty(
                key_type=types.intp,
                value_type=types.intp,
            )
            return d

        @njit
        def foo(count):
            d = Dict.empty(
                key_type=types.intp,
                value_type=inner_dict_ty,
            )
            for i in range(count):
                d[i] = inner_numba_dict()
                for j in range(i + 1):
                    d[i][j] = j

            return d

        d = foo(100)
        ct = 0
        for k, dd in d.items():
            ct += 1
            self.assertEqual(len(dd), k + 1)
            for kk, vv in dd.items():
                self.assertEqual(kk, vv)

        self.assertEqual(ct, 100)

    def test_delitem(self):
        d = Dict.empty(types.int64, types.unicode_type)
        d[1] = 'apple'

        @njit
        def foo(x, k):
            del x[1]

        foo(d, 1)
        self.assertEqual(len(d), 0)
        self.assertFalse(d)

    def test_getitem_return_type(self):
        # Dict.__getitem__ must return non-optional type.
        d = Dict.empty(types.int64, types.int64[:])
        d[1] = np.arange(10, dtype=np.int64)

        @njit
        def foo(d):
            d[1] += 100
            return d[1]

        foo(d)
        # Return type is an array, not optional
        retty = foo.nopython_signatures[0].return_type
        self.assertIsInstance(retty, types.Array)
        self.assertNotIsInstance(retty, types.Optional)
        # Value is correctly updated
        self.assertPreciseEqual(d[1], np.arange(10, dtype=np.int64) + 100)

    def test_storage_model_mismatch(self):
        # https://github.com/numba/numba/issues/4520
        # check for storage model mismatch in refcount ops generation
        dct = Dict()
        ref = [
            ("a", True, "a"),
            ("b", False, "b"),
            ("c", False, "c"),
        ]
        # populate
        for x in ref:
            dct[x] = x
        # test
        for i, x in enumerate(ref):
            self.assertEqual(dct[x], x)


class TestDictForbiddenTypes(TestCase):
    def assert_disallow(self, expect, callable):
        with self.assertRaises(TypingError) as raises:
            callable()
        msg = str(raises.exception)
        self.assertIn(expect, msg)

    def assert_disallow_key(self, ty):
        msg = '{} as key is forbidden'.format(ty)
        self.assert_disallow(msg, lambda: Dict.empty(ty, types.intp))

        @njit
        def foo():
            Dict.empty(ty, types.intp)
        self.assert_disallow(msg, foo)

    def assert_disallow_value(self, ty):
        msg = '{} as value is forbidden'.format(ty)
        self.assert_disallow(msg, lambda: Dict.empty(types.intp, ty))

        @njit
        def foo():
            Dict.empty(types.intp, ty)
        self.assert_disallow(msg, foo)

    def test_disallow_list(self):
        self.assert_disallow_key(types.List(types.intp))
        self.assert_disallow_value(types.List(types.intp))

    def test_disallow_set(self):
        self.assert_disallow_key(types.Set(types.intp))
        self.assert_disallow_value(types.Set(types.intp))


class TestDictInferred(TestCase):
    def test_simple_literal(self):
        @njit
        def foo():
            d = Dict()
            d[123] = 321
            return d

        k, v = 123, 321
        d = foo()
        self.assertEqual(dict(d), {k: v})
        self.assertEqual(typeof(d).key_type, typeof(k))
        self.assertEqual(typeof(d).value_type, typeof(v))

    def test_simple_args(self):
        @njit
        def foo(k, v):
            d = Dict()
            d[k] = v
            return d

        k, v = 123, 321
        d = foo(k, v)
        self.assertEqual(dict(d), {k: v})
        self.assertEqual(typeof(d).key_type, typeof(k))
        self.assertEqual(typeof(d).value_type, typeof(v))

    def test_simple_upcast(self):
        @njit
        def foo(k, v, w):
            d = Dict()
            d[k] = v
            d[k] = w
            return d

        k, v, w = 123, 32.1, 321
        d = foo(k, v, w)
        self.assertEqual(dict(d), {k: w})
        self.assertEqual(typeof(d).key_type, typeof(k))
        self.assertEqual(typeof(d).value_type, typeof(v))

    def test_conflicting_value_type(self):
        @njit
        def foo(k, v, w):
            d = Dict()
            d[k] = v
            d[k] = w
            return d

        k, v, w = 123, 321, 32.1
        with self.assertRaises(TypingError) as raises:
            foo(k, v, w)
        self.assertIn(
            'cannot safely cast float64 to {}'.format(typeof(v)),
            str(raises.exception),
        )

    def test_conflicting_key_type(self):
        @njit
        def foo(k, h, v):
            d = Dict()
            d[k] = v
            d[h] = v
            return d

        k, h, v = 123, 123.1, 321
        with self.assertRaises(TypingError) as raises:
            foo(k, h, v)
        self.assertIn(
            'cannot safely cast float64 to {}'.format(typeof(v)),
            str(raises.exception),
        )

    def test_conflict_key_type_non_number(self):
        # Allow non-number types to cast unsafely
        @njit
        def foo(k1, v1, k2):
            d = Dict()
            d[k1] = v1
            return d, d[k2]

        # k2 will unsafely downcast typeof(k1)
        k1 = (np.int8(1), np.int8(2))
        k2 = (np.int32(1), np.int32(2))
        v1 = np.intp(123)

        with warnings.catch_warnings(record=True) as w:
            d, dk2 = foo(k1, v1, k2)
        self.assertEqual(len(w), 1)
        # Make sure the warning is about unsafe cast
        msg = 'unsafe cast from UniTuple(int32 x 2) to UniTuple(int8 x 2)'
        self.assertIn(msg, str(w[0]))

        keys = list(d.keys())
        self.assertEqual(keys[0], (1, 2))
        self.assertEqual(dk2, d[(np.int32(1), np.int32(2))])

    def test_ifelse_filled_both_branches(self):
        @njit
        def foo(k, v):
            d = Dict()
            if k:
                d[k] = v
            else:
                d[0xdead] = v + 1

            return d

        k, v = 123, 321
        d = foo(k, v)
        self.assertEqual(dict(d), {k: v})
        k, v = 0, 0
        d = foo(k, v)
        self.assertEqual(dict(d), {0xdead: v + 1})

    def test_ifelse_empty_one_branch(self):
        @njit
        def foo(k, v):
            d = Dict()
            if k:
                d[k] = v

            return d

        k, v = 123, 321
        d = foo(k, v)
        self.assertEqual(dict(d), {k: v})
        k, v = 0, 0
        d = foo(k, v)
        self.assertEqual(dict(d), {})
        self.assertEqual(typeof(d).key_type, typeof(k))
        self.assertEqual(typeof(d).value_type, typeof(v))

    def test_loop(self):
        @njit
        def foo(ks, vs):
            d = Dict()
            for k, v in zip(ks, vs):
                d[k] = v
            return d

        vs = list(range(4))
        ks = list(map(lambda x : x + 100, vs))
        d = foo(ks, vs)
        self.assertEqual(dict(d), dict(zip(ks, vs)))

    def test_unused(self):
        @njit
        def foo():
            d = Dict()
            return d

        with self.assertRaises(TypingError) as raises:
            foo()
        self.assertIn(
            "imprecise type",
            str(raises.exception)
        )

    def test_define_after_use(self):
        @njit
        def foo(define):
            d = Dict()
            ct = len(d)
            for k, v in d.items():
                ct += v

            if define:
                # This will set the type
                d[1] = 2
            return ct, d, len(d)

        ct, d, n = foo(True)
        self.assertEqual(ct, 0)
        self.assertEqual(n, 1)
        self.assertEqual(dict(d), {1: 2})

        ct, d, n = foo(False)
        self.assertEqual(ct, 0)
        self.assertEqual(dict(d), {})
        self.assertEqual(n, 0)

    def test_dict_of_dict(self):
        @njit
        def foo(k1, k2, v):
            d = Dict()
            z1 = Dict()
            z1[k1 + 1] = v + k1
            z2 = Dict()
            z2[k2 + 2] = v + k2
            d[k1] = z1
            d[k2] = z2
            return d

        k1, k2, v = 100, 200, 321
        d = foo(k1, k2, v)
        self.assertEqual(
            dict(d),
            {
                k1: {k1 + 1: k1 + v},
                k2: {k2 + 2: k2 + v},
            },
        )

    def test_comprehension_basic(self):
        @njit
        def foo():
            return {i: 2 * i for i in range(10)}

        self.assertEqual(foo(), foo.py_func())

    def test_comprehension_basic_mixed_type(self):
        @njit
        def foo():
            return {i: float(j) for i, j in zip(range(10), range(10, 0, -1))}

        self.assertEqual(foo(), foo.py_func())

    def test_comprehension_involved(self):
        @njit
        def foo():
            a = {0: 'A', 1: 'B', 2: 'C'}
            return {3 + i: a[i] for i in range(3)}

        self.assertEqual(foo(), foo.py_func())

    def test_comprehension_fail_mixed_type(self):
        @njit
        def foo():
            a = {0: 'A', 1: 'B', 2: 1j}
            return {3 + i: a[i] for i in range(3)}

        with self.assertRaises(TypingError) as e:
            foo()

        excstr = str(e.exception)
        self.assertIn("Cannot cast complex128 to unicode_type", excstr)


class TestNonCompiledInfer(TestCase):
    def test_check_untyped_dict_ops(self):
        # Check operation on untyped dictionary
        d = Dict()
        self.assertFalse(d._typed)
        self.assertEqual(len(d), 0)
        self.assertEqual(str(d), str({}))
        self.assertEqual(list(iter(d)), [])
        # Test __getitem__
        with self.assertRaises(KeyError) as raises:
            d[1]
        self.assertEqual(str(raises.exception), str(KeyError(1)))
        # Test __delitem__
        with self.assertRaises(KeyError) as raises:
            del d[1]
        self.assertEqual(str(raises.exception), str(KeyError(1)))
        # Test .pop
        with self.assertRaises(KeyError):
            d.pop(1)
        self.assertEqual(str(raises.exception), str(KeyError(1)))
        # Test .pop
        self.assertIs(d.pop(1, None), None)
        # Test .get
        self.assertIs(d.get(1), None)
        # Test .popitem
        with self.assertRaises(KeyError) as raises:
            d.popitem()
        self.assertEqual(str(raises.exception),
                         str(KeyError('dictionary is empty')))
        # Test setdefault(k)
        with self.assertRaises(TypeError) as raises:
            d.setdefault(1)
        self.assertEqual(
            str(raises.exception),
            str(TypeError('invalid operation on untyped dictionary')),
        )
        # Test __contains__
        self.assertFalse(1 in d)
        # It's untyped
        self.assertFalse(d._typed)

    def test_getitem(self):
        # Test __getitem__
        d = Dict()
        d[1] = 2
        # It's typed now
        self.assertTrue(d._typed)
        self.assertEqual(d[1], 2)

    def test_setdefault(self):
        # Test setdefault(k, d)
        d = Dict()
        d.setdefault(1, 2)
        # It's typed now
        self.assertTrue(d._typed)
        self.assertEqual(d[1], 2)


@jitclass(spec=[('a', types.intp)])
class Bag(object):
    def __init__(self, a):
        self.a = a

    def __hash__(self):
        return hash(self.a)


class TestDictWithJitclass(TestCase):
    def test_jitclass_as_value(self):
        @njit
        def foo(x):
            d = Dict()
            d[0] = x
            d[1] = Bag(101)
            return d

        d = foo(Bag(a=100))
        self.assertEqual(d[0].a, 100)
        self.assertEqual(d[1].a, 101)


class TestNoJit(TestCase):
    """Exercise dictionary creation with JIT disabled. """

    def test_dict_create_no_jit_using_new_dict(self):
        with override_config('DISABLE_JIT', True):
            with forbid_codegen():
                d = dictobject.new_dict(int32, float32)
                self.assertEqual(type(d), dict)

    def test_dict_create_no_jit_using_Dict(self):
        with override_config('DISABLE_JIT', True):
            with forbid_codegen():
                d = Dict()
                self.assertEqual(type(d), dict)

    def test_dict_create_no_jit_using_empty(self):
        with override_config('DISABLE_JIT', True):
            with forbid_codegen():
                d = Dict.empty(types.int32, types.float32)
                self.assertEqual(type(d), dict)


class TestDictIterator(TestCase):
    def test_dict_iterator(self):
        @njit
        def fun1():
            dd = Dict.empty(key_type=types.intp,
                            value_type=types.intp)
            dd[0] = 10
            dd[1] = 20
            dd[2] = 30

            return list(dd.keys()), list(dd.values())

        @njit
        def fun2():
            dd = Dict.empty(key_type=types.intp,
                            value_type=types.intp)
            dd[4] = 77
            dd[5] = 88
            dd[6] = 99

            return list(dd.keys()), list(dd.values())
        res1 = fun1()
        res2 = fun2()

        self.assertEqual([0,1,2], res1[0])
        self.assertEqual([10,20,30], res1[1])
        self.assertEqual([4,5,6], res2[0])
        self.assertEqual([77,88,99], res2[1])


class TestTypedDictInitialValues(MemoryLeakMixin, TestCase):
    """Tests that typed dictionaries carry their initial value if present"""

    def test_homogeneous_and_literal(self):
        def bar(d):
            ...

        @overload(bar)
        def ol_bar(d):
            if d.initial_value is None:
                return lambda d: literally(d)
            self.assertTrue(isinstance(d, types.DictType))
            self.assertEqual(d.initial_value, {'a': 1, 'b': 2, 'c': 3})
            self.assertEqual(hasattr(d, 'literal_value'), False)
            return lambda d: d

        @njit
        def foo():
            # keys and values all have literal representation
            x = {'a': 1, 'b': 2, 'c': 3}
            bar(x)

        foo()

    def test_heterogeneous_but_castable_to_homogeneous(self):
        def bar(d):
            ...

        @overload(bar)
        def ol_bar(d):
            self.assertTrue(isinstance(d, types.DictType))
            self.assertEqual(d.initial_value, None)
            self.assertEqual(hasattr(d, 'literal_value'), False)
            return lambda d: d

        @njit
        def foo():
            # This dictionary will be typed based on 1j, i.e. complex128
            # as the values are not all literals, there's no "initial_value"
            # available irrespective of whether it's possible to rip this
            # information out of the bytecode.
            x = {'a': 1j, 'b': 2, 'c': 3}
            bar(x)

        foo()

    def test_heterogeneous_but_not_castable_to_homogeneous(self):
        def bar(d):
            ...

        @overload(bar)
        def ol_bar(d):
            a = {'a': 1, 'b': 2j, 'c': 3}

            def specific_ty(z):
                return types.literal(z) if types.maybe_literal(z) else typeof(z)
            expected = {types.literal(x): specific_ty(y) for x, y in a.items()}
            self.assertTrue(isinstance(d, types.LiteralStrKeyDict))
            self.assertEqual(d.literal_value, expected)
            self.assertEqual(hasattr(d, 'initial_value'), False)
            return lambda d: d

        @njit
        def foo():
            # This dictionary will be typed based on 1, i.e. intp, as the values
            # cannot all be cast to this type, but the keys are literal strings
            # this is a LiteralStrKey[Dict], there's no initial_value but there
            # is a literal_value.
            x = {'a': 1, 'b': 2j, 'c': 3}
            bar(x)

        foo()

    def test_mutation_not_carried(self):
        def bar(d):
            ...

        @overload(bar)
        def ol_bar(d):
            if d.initial_value is None:
                return lambda d: literally(d)
            self.assertTrue(isinstance(d, types.DictType))
            self.assertEqual(d.initial_value, {'a': 1, 'b': 2, 'c': 3})
            return lambda d: d

        @njit
        def foo():
            # This dictionary is mutated, check the initial_value carries
            # correctly and is not mutated
            x = {'a': 1, 'b': 2, 'c': 3}
            x['d'] = 4
            bar(x)

        foo()

    def test_mutation_not_carried_single_function(self):
        # this is another pattern for using literally

        @njit
        def nop(*args):
            pass

        for fn, iv in (nop, None), (literally, {'a': 1, 'b': 2, 'c': 3}):
            @njit
            def baz(x):
                pass

            def bar(z):
                pass

            @overload(bar)
            def ol_bar(z):
                def impl(z):
                    fn(z)
                    baz(z)
                return impl

            @njit
            def foo():
                x = {'a': 1, 'b': 2, 'c': 3}
                bar(x)
                x['d'] = 4
                return x

            foo()
            # baz should be specialised based on literally being invoked and
            # the literal/unliteral arriving at the call site
            larg = baz.signatures[0][0]
            self.assertEqual(larg.initial_value, iv)

    def test_unify_across_function_call(self):

        @njit
        def bar(x):
            o = {1: 2}
            if x:
                o = {2: 3}
            return o

        @njit
        def foo(x):
            if x:
                d = {3: 4}
            else:
                d = bar(x)
            return d

        e1 = Dict()
        e1[3] = 4
        e2 = Dict()
        e2[1] = 2
        self.assertEqual(foo(True), e1)
        self.assertEqual(foo(False), e2)


class TestLiteralStrKeyDict(MemoryLeakMixin, TestCase):
    """ Tests for dictionaries with string keys that can map to anything!"""

    def test_basic_const_lowering_boxing(self):
        @njit
        def foo():
            ld = {'a': 1, 'b': 2j, 'c': 'd'}
            return (ld['a'], ld['b'], ld['c'])

        self.assertEqual(foo(), (1, 2j, 'd'))

    def test_basic_nonconst_in_scope(self):
        @njit
        def foo(x):
            y = x + 5
            e = True if y > 2 else False
            ld = {'a': 1, 'b': 2j, 'c': 'd', 'non_const': e}
            return ld['non_const']

        # Recall that key non_const has a value of a known type, bool, and it's
        # value is stuffed in at run time, this is permitted as the dictionary
        # is immutable in type
        self.assertTrue(foo(34))
        self.assertFalse(foo(-100))

    def test_basic_nonconst_freevar(self):
        e = 5

        def bar(x):
            pass

        @overload(bar)
        def ol_bar(x):
            self.assertEqual(x.literal_value,
                             {types.literal('a'): types.literal(1),
                              types.literal('b'): typeof(2j),
                              types.literal('c'): types.literal('d'),
                              types.literal('d'): types.literal(5)})

            def impl(x):
                pass
            return impl

        @njit
        def foo():
            ld = {'a': 1, 'b': 2j, 'c': 'd', 'd': e}
            bar(ld)

        foo()

    def test_literal_value(self):

        def bar(x):
            pass

        @overload(bar)
        def ol_bar(x):
            self.assertEqual(x.literal_value,
                             {types.literal('a'): types.literal(1),
                              types.literal('b'): typeof(2j),
                              types.literal('c'): types.literal('d')})

            def impl(x):
                pass
            return impl

        @njit
        def foo():
            ld = {'a': 1, 'b': 2j, 'c': 'd'}
            bar(ld)

        foo()

    def test_list_and_array_as_value(self):

        def bar(x):
            pass

        @overload(bar)
        def ol_bar(x):
            self.assertEqual(x.literal_value,
                             {types.literal('a'): types.literal(1),
                              types.literal('b'):
                              types.List(types.intp, initial_value=[1,2,3]),
                              types.literal('c'): typeof(np.zeros(5))})

            def impl(x):
                pass
            return impl

        @njit
        def foo():
            b = [1, 2, 3]
            ld = {'a': 1, 'b': b, 'c': np.zeros(5)}
            bar(ld)

        foo()

    def test_repeated_key_literal_value(self):

        def bar(x):
            pass

        @overload(bar)
        def ol_bar(x):
            # order is important, 'a' was seen first, but updated later
            self.assertEqual(x.literal_value,
                             {types.literal('a'): types.literal('aaaa'),
                              types.literal('b'): typeof(2j),
                              types.literal('c'): types.literal('d')})

            def impl(x):
                pass
            return impl

        @njit
        def foo():
            ld = {'a': 1, 'a': 10, 'b': 2j, 'c': 'd', 'a': 'aaaa'} # noqa #F601
            bar(ld)

        foo()

    def test_read_only(self):

        def _len():
            ld = {'a': 1, 'b': 2j, 'c': 'd'}
            return len(ld)

        def static_getitem():
            ld = {'a': 1, 'b': 2j, 'c': 'd'}
            return ld['b']

        def contains():
            ld = {'a': 1, 'b': 2j, 'c': 'd'}
            return 'b' in ld, 'f' in ld

        def copy():
            ld = {'a': 1, 'b': 2j, 'c': 'd'}
            new = ld.copy()
            return ld == new

        rdonlys = (_len, static_getitem, contains, copy)

        for test in rdonlys:
            with self.subTest(test.__name__):
                self.assertPreciseEqual(njit(test)(), test())

    def test_mutation_failure(self):

        def setitem():
            ld = {'a': 1, 'b': 2j, 'c': 'd'}
            ld['a'] = 12

        def delitem():
            ld = {'a': 1, 'b': 2j, 'c': 'd'}
            del ld['a']

        def popitem():
            ld = {'a': 1, 'b': 2j, 'c': 'd'}
            ld.popitem()

        def pop():
            ld = {'a': 1, 'b': 2j, 'c': 'd'}
            ld.pop()

        def clear():
            ld = {'a': 1, 'b': 2j, 'c': 'd'}
            ld.clear()

        def setdefault():
            ld = {'a': 1, 'b': 2j, 'c': 'd'}
            ld.setdefault('f', 1)

        illegals = (setitem, delitem, popitem, pop, clear, setdefault)

        for test in illegals:
            with self.subTest(test.__name__):
                with self.assertRaises(TypingError) as raises:
                    njit(test)()
                expect = "Cannot mutate a literal dictionary"
                self.assertIn(expect, str(raises.exception))

    def test_get(self):

        @njit
        def get(x):
            ld = {'a': 2j, 'c': 'd'}
            return ld.get(x)

        @njit
        def getitem(x):
            ld = {'a': 2j, 'c': 'd'}
            return ld[x]

        for test in (get, getitem):
            with self.subTest(test.__name__):
                with self.assertRaises(TypingError) as raises:
                    test('a')
                expect = "Cannot get{item}() on a literal dictionary"
                self.assertIn(expect, str(raises.exception))

    def test_dict_keys(self):

        @njit
        def foo():
            ld = {'a': 2j, 'c': 'd'}
            return [x for x in ld.keys()]

        self.assertEqual(foo(), ['a', 'c'])

    def test_dict_values(self):

        @njit
        def foo():
            ld = {'a': 2j, 'c': 'd'}
            return ld.values()

        self.assertEqual(foo(), (2j, 'd'))

    def test_dict_items(self):
        @njit
        def foo():
            ld = {'a': 2j, 'c': 'd', 'f': np.zeros((5))}
            return ld.items()

        self.assertPreciseEqual(foo(),
                                (('a', 2j), ('c', 'd'), ('f', np.zeros((5)))))

    def test_dict_return(self):

        @njit
        def foo():
            ld = {'a': 2j, 'c': 'd'}
            return ld

        # escaping heterogeneous dictionary is not supported
        with self.assertRaises(TypeError) as raises:
            foo()

        excstr = str(raises.exception)
        self.assertIn("cannot convert native LiteralStrKey", excstr)

    def test_dict_unify(self):
        @njit
        def foo(x):
            if x + 7 > 4:
                a = {'a': 2j, 'c': 'd', 'e': np.zeros(4)}
            else:
                # Note the use of a different literal str for key 'c'
                a = {'a': 5j, 'c': 'CAT', 'e': np.zeros((5,))}
            return a['c']

        self.assertEqual(foo(100), 'd')
        self.assertEqual(foo(-100), 'CAT')
        self.assertEqual(foo(100), foo.py_func(100))
        self.assertEqual(foo(-100), foo.py_func(-100))

    def test_dict_not_unify(self):

        @njit
        def key_mismatch(x):
            if x + 7 > 4:
                a = {'BAD_KEY': 2j, 'c': 'd', 'e': np.zeros(4)}
            else:
                a = {'a': 5j, 'c': 'CAT', 'e': np.zeros((5,))}
            # prevents inline of return on py310
            py310_defeat1 = 1  # noqa
            py310_defeat2 = 2  # noqa
            py310_defeat3 = 3  # noqa
            py310_defeat4 = 4  # noqa
            return a['a']

        with self.assertRaises(TypingError) as raises:
            key_mismatch(100)

        self.assertIn("Cannot unify LiteralStrKey", str(raises.exception))

        @njit
        def value_type_mismatch(x):
            if x + 7 > 4:
                a = {'a': 2j, 'c': 'd', 'e': np.zeros((4, 3))}
            else:
                a = {'a': 5j, 'c': 'CAT', 'e': np.zeros((5,))}
            # prevents inline of return on py310
            py310_defeat1 = 1  # noqa
            py310_defeat2 = 2  # noqa
            py310_defeat3 = 3  # noqa
            py310_defeat4 = 4  # noqa
            return a['a']

        with self.assertRaises(TypingError) as raises:
            value_type_mismatch(100)

        self.assertIn("Cannot unify LiteralStrKey", str(raises.exception))

    def test_dict_value_coercion(self):
        # checks that things coerce or not!

        p = {# safe and no conversion: TypedDict
             (np.int32, np.int32): types.DictType,
             # safe and convertible: TypedDict
             (np.int32, np.int8): types.DictType,
             # safe convertible: TypedDict
             (np.complex128, np.int32): types.DictType,
             # unsafe not convertible: LiteralStrKey
             (np.int32, np.complex128): types.LiteralStrKeyDict,
             # unsafe not convertible: LiteralStrKey
             (np.int32, np.array): types.LiteralStrKeyDict,
             # unsafe not convertible: LiteralStrKey
             (np.array, np.int32): types.LiteralStrKeyDict,
             # unsafe not convertible: LiteralStrKey
             (np.int8, np.int32): types.LiteralStrKeyDict,
             # unsafe not convertible: LiteralStrKey (issue #6420 case)
             (np.int64, np.float64): types.LiteralStrKeyDict,}

        def bar(x):
            pass

        for dts, container in p.items():
            @overload(bar)
            def ol_bar(x):
                self.assertTrue(isinstance(x, container))

                def impl(x):
                    pass
                return impl

            ty1, ty2 = dts

            @njit
            def foo():
                d = {'a': ty1(1), 'b': ty2(2)}
                bar(d)

            foo()

    def test_build_map_op_code(self):
        # tests building dictionaries via `build_map`, which, for statically
        # determinable str key->things cases is just a single key:value
        # any other build_map would either end up as being non-const str keys
        # or keys of some non-string type and therefore not considered.
        def bar(x):
            pass

        @overload(bar)
        def ol_bar(x):
            def impl(x):
                pass
            return impl

        @njit
        def foo():
            a = {'a': {'b1': 10, 'b2': 'string'}}
            bar(a)

        foo()

    def test_dict_as_arg(self):
        @njit
        def bar(fake_kwargs=None):
            if fake_kwargs is not None:
                # Add 10 to array in key 'd'
                fake_kwargs['d'][:] += 10

        @njit
        def foo():
            a = 1
            b = 2j
            c = 'string'
            d = np.zeros(3)
            e = {'a': a, 'b': b, 'c': c, 'd': d}
            bar(fake_kwargs=e)
            return e['d']

        np.testing.assert_allclose(foo(), np.ones(3) * 10)

    def test_dict_with_single_literallist_value(self):
        #see issue #6094
        @njit
        def foo():
            z = {"A": [lambda a: 2 * a, "B"]}
            return z["A"][0](5)

        self.assertPreciseEqual(foo(), foo.py_func())

    def test_tuple_not_in_mro(self):
        # Related to #6094, make sure that LiteralStrKey does not inherit from
        # types.BaseTuple as this breaks isinstance checks.
        def bar(x):
            pass

        @overload(bar)
        def ol_bar(x):
            self.assertFalse(isinstance(x, types.BaseTuple))
            self.assertTrue(isinstance(x, types.LiteralStrKeyDict))
            return lambda x: ...

        @njit
        def foo():
            d = {'a': 1, 'b': 'c'}
            bar(d)

        foo()

    def test_const_key_not_in_dict(self):

        @njit
        def foo():
            a = {'not_a': 2j, 'c': 'd', 'e': np.zeros(4)}
            return a['a']

        with self.assertRaises(TypingError) as raises:
            foo()

        self.assertIn("Key 'a' is not in dict.", str(raises.exception))

    def test_uncommon_identifiers(self):
        # Tests uncommon identifiers like numerical values and operators in
        # the key fields. See #6518 and #7416.

        # Numerical values in keys
        @njit
        def foo():
            d = {'0': np.ones(5), '1': 4}
            return len(d)

        self.assertPreciseEqual(foo(), foo.py_func())

        # operators in keys
        @njit
        def bar():
            d = {'+': np.ones(5), 'x--': 4}
            return len(d)

        self.assertPreciseEqual(bar(), bar.py_func())

    def test_update_error(self):
        # Tests that dict.update produces a reasonable
        # error with a LiteralStrKeyDict input.
        @njit
        def foo():

            d1 = {
                'a': 2,
                'b': 4,
                'c': 'a'
            }
            d1.update({'x': 3})
            return d1

        with self.assertRaises(TypingError) as raises:
            foo()

        self.assertIn(
            "Cannot mutate a literal dictionary",
            str(raises.exception)
        )


if __name__ == '__main__':
    unittest.main()
